#ifndef _PCIP_
#define _PCIP_

typedef enum {
    BaseClassDescription,
    SubClassDescription,
    ProgIfDescription
} PCI_CLASSCODEDESC;


typedef struct _PCI_BRIDBG_CTRL_REG {
    USHORT PERRREnable:1;
    USHORT SERREnable:1;
    USHORT ISAEnable:1;
    USHORT VGAEnable:1;

    USHORT Reserved:1;
    USHORT MasterAbort:1;
    USHORT CBRst:1;
    USHORT IRQRoutingEnable:1;

    USHORT Mem0Prefetch:1;
    USHORT Mem1Prefetch:1;
    USHORT WritePostEnable:1;

    USHORT Reserved1:5;
} PCI_BRIDBG_CTRL_REG;

//
// Cardbus has extra configuration information beyond the common
// header. (stolen from the PCI driver)
//
typedef struct _PCI_TYPE2_HEADER_EXTRAS {
    USHORT  SubVendorID;
    USHORT  SubSystemID;
    ULONG   LegacyModeBaseAddress;
} PCI_TYPE2_HEADER_EXTRAS, *PPCI_TYPE2_HEADER_EXTRAS;

#define CONFIG_OFFSET(x)    \
        FIELD_OFFSET(PCI_COMMON_CONFIG, x)

#define TYPE0_OFFSET(x)     \
        CONFIG_OFFSET(u.type0.x)
        
#define TYPE1_OFFSET(x)     \
        CONFIG_OFFSET(u.type1.x)
        
#define TYPE2_OFFSET(x)     \
        CONFIG_OFFSET(u.type2.x)
        
VOID
PrintPciHtCommandReg(
    IN PPCI_HT_CAPABILITY PciHtCap
    );

VOID
DumpCfgSpace (
    IN PPCI_TYPE1_CFG_BITS PciCfg1
    );

PCHAR 
GetClassDesc(
    IN UCHAR BaseClass, 
    IN UCHAR SubClass, 
    IN UCHAR ProgIf,
    IN PCI_CLASSCODEDESC ClassDesc
    );

PCHAR
GetVendorDesc(
    IN USHORT VendorID,
    IN BOOL FullVenDesc
    );

BOOL
ReadPci (
    IN PPCI_TYPE1_CFG_BITS      PciCfg1,
    OUT PUCHAR                  Buffer,
    IN ULONG                    Offset,
    IN ULONG                    Length
    );

BOOLEAN
WritePci (
    IN PPCI_TYPE1_CFG_BITS      PciCfg1,
    IN PUCHAR                   Buffer,
    IN ULONG                    Offset,
    IN ULONG                    Length
    );

static struct _PCI_CLASSCODETABLE{
	UCHAR BaseClass;
	UCHAR SubClass;
	UCHAR ProgIf;
	PCHAR BaseDesc;
	PCHAR SubDesc;
	PCHAR ProgDesc;   
} PciClassCodeTable[] = {

	{ 0x00, 0x00, 0x00, "Pre-2.0 PCI Specification Device", "Non-VGA",NULL } ,
	{ 0x00, 0x01, 0x00, "Pre-2.0 PCI Specification Device", "VGA Compatible", NULL } ,

	{ 0x01, 0x00, 0x00, "Mass Storage Controller", "SCSI Controller", NULL } ,
	{ 0x01, 0x01, 0x00, "Mass Storage Controller", "IDE Controller", NULL } ,
	{ 0x01, 0x02, 0x00, "Mass Storage Controller", "Floppy Disk Controller", NULL } ,
	{ 0x01, 0x03, 0x00, "Mass Storage Controller", "IPI Controller", NULL } ,
	{ 0x01, 0x04, 0x00, "Mass Storage Controller", "RAID Controller", NULL } ,
	{ 0x01, 0x80, 0x00, "Mass Storage Controller", "Other Storage Controller", NULL } ,

	{ 0x02, 0x00, 0x00, "Network Controller", "Ethernet Controller", NULL } ,
	{ 0x02, 0x01, 0x00, "Network Controller", "Token Ring Controller", NULL } ,
	{ 0x02, 0x02, 0x00, "Network Controller", "FDDI Controller", NULL } ,
	{ 0x02, 0x03, 0x00, "Network Controller", "ATM Controller", NULL } ,
	{ 0x02, 0x04, 0x00, "Network Controller", "ISDN Controller", NULL } ,
	{ 0x02, 0x80, 0x00, "Network Controller", "Other Network Controller", NULL } ,

	{ 0x03, 0x00, 0x00, "Display Controller", "VGA Compatible Controller", "VGA" } ,
	{ 0x03, 0x00, 0x01, "Display Controller", "VGA Compatible Controller", "8514" } ,
	{ 0x03, 0x01, 0x00, "Display Controller", "XGA Controller", NULL } ,
	{ 0x03, 0x02, 0x00, "Display Controller", "3D Controller", NULL } ,
	{ 0x03, 0x80, 0x00, "Display Controller", "Other Display Controller", NULL } ,

	{ 0x04, 0x00, 0x00, "Multimedia Device", "Video Device", NULL } ,
	{ 0x04, 0x01, 0x00, "Multimedia Device", "Audio Device", NULL } ,
	{ 0x04, 0x02, 0x00, "Multimedia Device", "Telephony Device", NULL } ,
	{ 0x04, 0x80, 0x00, "Multimedia Device", "Other Device", NULL } ,

	{ 0x05, 0x00, 0x00, "Memory Controller", "RAM Controller", NULL } ,
	{ 0x05, 0x01, 0x00, "Memory Controller", "Flash Controller", NULL } ,
	{ 0x05, 0x80, 0x00, "Memory Controller", "Other Memory Controller", NULL } ,

	{ 0x06, 0x00, 0x00, "Bridge Device", "Host Bridge", NULL } ,
	{ 0x06, 0x01, 0x00, "Bridge Device", "ISA Bridge", NULL } ,
	{ 0x06, 0x02, 0x00, "Bridge Device", "EISA Bridge", NULL } ,
	{ 0x06, 0x03, 0x00, "Bridge Device", "MCA Bridge", NULL } ,
	{ 0x06, 0x04, 0x00, "Bridge Device", "PCI-PCI Bridge", NULL } ,
	{ 0x06, 0x04, 0x01, "Bridge Device", "PCI-PCI Bridge", "Subtractive" } ,
	{ 0x06, 0x05, 0x00, "Bridge Device", "PCMCIA Bridge", NULL } ,
	{ 0x06, 0x06, 0x00, "Bridge Device", "NuBus Bridge", NULL } ,
	{ 0x06, 0x07, 0x00, "Bridge Device", "CardBus Bridge", NULL } ,
	{ 0x06, 0x08, 0x00, "Bridge Device", "RACEway Bridge", "Transparent" } ,
	{ 0x06, 0x08, 0x01, "Bridge Device", "RACEway Bridge", "End-Point" } ,
	{ 0x06, 0x80, 0x00, "Bridge Device", "Other Bridge", NULL } ,

	{ 0x07, 0x00, 0x00, "Simple Communications Controller", "Serial Controller", "Generic XT Compatible" } ,
	{ 0x07, 0x00, 0x01, "Simple Communications Controller", "Serial Controller", "16450 Compatible" } ,
	{ 0x07, 0x00, 0x02, "Simple Communications Controller", "Serial Controller", "16550 Compatible" } ,
	{ 0x07, 0x00, 0x03, "Simple Communications Controller", "Serial Controller", "16650 Compatible" } ,
	{ 0x07, 0x00, 0x04, "Simple Communications Controller", "Serial Controller", "16750 Compatible" } ,
	{ 0x07, 0x00, 0x05, "Simple Communications Controller", "Serial Controller", "16850 Compatible" } ,
	{ 0x07, 0x00, 0x06, "Simple Communications Controller", "Serial Controller", "16950 Compatible" } ,
	{ 0x07, 0x01, 0x00, "Simple Communications Controller", "Parallel Port", "Standard" } ,
	{ 0x07, 0x01, 0x01, "Simple Communications Controller", "Parallel Port", "Bidirectional" } ,
	{ 0x07, 0x01, 0x02, "Simple Communications Controller", "Parallel Port", "ECP 1.X Compliant" } ,
	{ 0x07, 0x01, 0x03, "Simple Communications Controller", "Parallel Port", "IEEE 1284 controller" } ,
	{ 0x07, 0x01, 0xFE, "Simple Communications Controller", "Parallel Port", "IEEE 1284 target" } ,
	{ 0x07, 0x02, 0x00, "Simple Communications Controller", "Multiport Serial Controller", NULL } ,
	{ 0x07, 0x03, 0x00, "Simple Communications Controller", "Generic Modem", NULL } ,
	{ 0x07, 0x03, 0x01, "Simple Communications Controller", "Hayes-Compatible Modem", "16450" } ,
	{ 0x07, 0x03, 0x02, "Simple Communications Controller", "Hayes-Compatible Modem", "16550" } ,
	{ 0x07, 0x03, 0x03, "Simple Communications Controller", "Hayes-Compatible Modem", "16650" } ,
	{ 0x07, 0x03, 0x04, "Simple Communications Controller", "Hayes-Compatible Modem", "16750" } ,
	{ 0x07, 0x80, 0x00, "Simple Communications Controller", "Other Communications Controller", NULL } ,

	{ 0x08, 0x00, 0x00, "Base Systems Peripheral", "Interrupt Controller", "Generic 8259" } ,
	{ 0x08, 0x00, 0x01, "Base Systems Peripheral", "Interrupt Controller", "ISA" } ,
	{ 0x08, 0x00, 0x02, "Base Systems Peripheral", "Interrupt Controller", "EISA" } ,
	{ 0x08, 0x00, 0x10, "Base Systems Peripheral", "Interrupt Controller", "IO APIC" } ,
	{ 0x08, 0x00, 0x20, "Base Systems Peripheral", "Interrupt Controller", "IO(x) APIC" } ,
	{ 0x08, 0x01, 0x00, "Base Systems Peripheral", "DMA Controller", "Generic 8237" } ,
	{ 0x08, 0x01, 0x01, "Base Systems Peripheral", "DMA Controller", "ISA" } ,
	{ 0x08, 0x01, 0x02, "Base Systems Peripheral", "DMA Controller", "EISA" } ,
	{ 0x08, 0x02, 0x00, "Base Systems Peripheral", "System Timer", "Generic 8254" } ,
	{ 0x08, 0x02, 0x01, "Base Systems Peripheral", "System Timer", "ISA" } ,
	{ 0x08, 0x02, 0x02, "Base Systems Peripheral", "System Timer", "EISA" } ,
	{ 0x08, 0x03, 0x00, "Base Systems Peripheral", "RTC (Real Time Clock)", "Generic" } ,
	{ 0x08, 0x03, 0x01, "Base Systems Peripheral", "RTC (Real Time Clock)", "ISA" } ,
	{ 0x08, 0x04, 0x00, "Base Systems Peripheral", "PCI Hot-Plug Controller", "Generic" } ,
	{ 0x08, 0x80, 0x00, "Base Systems Peripheral", "Other System Peripheral", NULL } ,

	{ 0x09, 0x00, 0x00, "Input Device", "Keyboard Controller", NULL } ,
	{ 0x09, 0x01, 0x00, "Input Device", "Digitizer (Pen)", NULL } ,
	{ 0x09, 0x02, 0x00, "Input Device", "Mouse Controller", NULL } ,
	{ 0x09, 0x03, 0x00, "Input Device", "Scanner Controller", NULL } ,
	{ 0x09, 0x04, 0x00, "Input Device", "Game Port Controller", "Generic" } ,
	{ 0x09, 0x04, 0x10, "Input Device", "Game Port Controller", "Legacy" } ,
	{ 0x09, 0x80, 0x00, "Input Device", "Other Input Controller", NULL } ,

	{ 0x0A, 0x00, 0x00, "Docking Station", "Generic", NULL } ,
	{ 0x0A, 0x80, 0x00, "Docking Station", "Other", NULL } ,

	{ 0x0B, 0x00, 0x00, "Processor", "386", NULL } ,
	{ 0x0B, 0x01, 0x00, "Processor", "486", NULL } ,
	{ 0x0B, 0x02, 0x00, "Processor", "Pentium", NULL } ,
	{ 0x0B, 0x10, 0x00, "Processor", "Alpha", NULL } ,
	{ 0x0B, 0x20, 0x00, "Processor", "Power PC", NULL } ,
	{ 0x0B, 0x30, 0x00, "Processor", "MIPS", NULL } ,
	{ 0x0B, 0x40, 0x00, "Processor", "Co-processor", NULL } ,

	{ 0x0C, 0x00, 0x00, "Serial Bus Controller", "IEEE 1394 Controller", "Firewire" } ,
	{ 0x0C, 0x00, 0x10, "Serial Bus Controller", "IEEE 1394 Controller", "OpenHCI" } ,
	{ 0x0C, 0x01, 0x00, "Serial Bus Controller", "ACCESS.bus", NULL } ,
	{ 0x0C, 0x02, 0x00, "Serial Bus Controller", "SSA (Serial Storage Archetecture)", NULL } ,
	{ 0x0C, 0x03, 0x00, "Serial Bus Controller", "USB Controller", "Universal Host Controller" } ,
	{ 0x0C, 0x03, 0x10, "Serial Bus Controller", "USB Controller", "Open Host Controller" } ,
    { 0x0C, 0x03, 0x20, "Serial Bus Controller", "USB2 Controller", "USB2 Enhanced Interface" } ,
	{ 0x0C, 0x03, 0x80, "Serial Bus Controller", "USB Controller", "Non-specific Controller" } ,
	{ 0x0C, 0x03, 0xFE, "Serial Bus Controller", "USB Controller", "Device" } ,
	{ 0x0C, 0x04, 0x00, "Serial Bus Controller", "Fibre Channel", NULL } ,
	{ 0x0C, 0x05, 0x00, "Serial Bus Controller", "SMBus Controller", NULL } ,

	{ 0x0D, 0x00, 0x00, "Wireless", "iRDA-compatible Controller", NULL } ,
	{ 0x0D, 0x01, 0x00, "Wireless", "Consumer IR Controller", NULL } ,
	{ 0x0D, 0x10, 0x00, "Wireless", "RF Controller", NULL } ,
	{ 0x0D, 0x80, 0x00, "Wireless", "Other Wireless Controller", NULL } ,

	{ 0x0E, 0x00, 0x00, "Intelligent IO Controller", "I2O Controller", NULL } ,

	{ 0x0F, 0x01, 0x00, "Satellite Communications Controller", "TV Sat Controller", NULL } ,
	{ 0x0F, 0x02, 0x00, "Satellite Communications Controller", "Audio Sat Controller", NULL } ,
	{ 0x0F, 0x03, 0x00, "Satellite Communications Controller", "Voice Sat Controller", NULL } ,
	{ 0x0F, 0x04, 0x00, "Satellite Communications Controller", "Data Sat Controller", NULL } ,

	{ 0x10, 0x00, 0x00, "Encryption Controller", "Network/Computing Encryption Controller", NULL } ,
	{ 0x10, 0x10, 0x00, "Encryption Controller", "Entertainment Encryption Controller", NULL } ,
	{ 0x10, 0x80, 0x00, "Encryption Controller", "Other Encryption Controller", NULL } ,

	{ 0x11, 0x00, 0x00, "Signal Processing", "DPIO  Controller", NULL } ,
	{ 0x11, 0x80, 0x00, "Signal Processing", "Other Signal Processing Controller", NULL } ,

	{ 0xFF, 0xFF, 0xFF, "Unknown", "Unknown", "Unknown" } ,
    { 0x00, 0x00, 0x00, NULL, NULL, NULL}
};

static struct _PCI_VENTABLE{
    USHORT VenId;
    PCHAR VenShort;
    PCHAR VenFull;
} PciVenTable[] = {
    { 0x0000, "Gammagraphx", "Gammagraphx, Inc." } ,
    { 0x001A, "Ascend", "Ascend Communications, Inc." } ,
    { 0x0033, NULL, "Paradyne Corp." } ,
    { 0x003D, "Lockheed Martin", "Lockheed Martin Corp" } ,
    { 0x0070, "Hauppauge", "Hauppauge Computer Works Inc." } ,
    { 0x0100, NULL, "Ncipher Corp. Ltd" } ,
    { 0x0123, NULL, "General Dynamics" } ,
    { 0x0160, "USR", "us robotics" } ,
    { 0x0675, "Dynalink", "Dynalink" } ,
    { 0x0871, "Berkom", "Berkom" } ,
    { 0x0A89, "BREA", "BREA Technologies Inc." } ,
    { 0x0E11, "Compaq", "Compaq Computer Corp." } ,
    { 0x1000, "LSI", "LSI Logic" } ,
    { 0x1001, "KOLTER", "Kolter Electronic - Germany" } ,
    { 0x1002, "ATI", "ATI Technologies" } ,
    { 0x1003, "ULSI", "ULSI" } ,
    { 0x1004, "VLSI", "VLSI Technology" } ,
    { 0x1005, "Avance", "Avance Logic Inc." } ,
    { 0x1006, "Reply", "Reply Group" } ,
    { 0x1007, "NetFrame", "Netframe Systems" } ,
    { 0x1008, "Epson", "Epson" } ,
    { 0x100A, "Phoenix", "Phoenix Technologies Ltd." } ,
    { 0x100B, "NSC", "National Semiconductor" } ,
    { 0x100C, "Tseng", "Tseng Labs" } ,
    { 0x100D, "AST", "AST Research" } ,
    { 0x100E, "Weitek", "Weitek" } ,
    { 0x1010, "VLogic", "Video Logic Ltd." } ,
    { 0x1011, "DEC", "Digital Equipment Corporation" } ,
    { 0x1012, "Micronics", "Micronics Computers Inc." } ,
    { 0x1013, "Cirrus", "Cirrus Logic" } ,
    { 0x1014, "IBM", "International Business Machines Corp." } ,
    { 0x1015, "LSIL", "LSI Logic Corp of Canada" } ,
    { 0x1016, "Fujitsu ICL", "Fujitsu ICL Computers" } ,
    { 0x1017, "Spea", "Spea Software AG" } ,
    { 0x1018, "Unisys", "Unisys Systems" } ,
    { 0x1019, "ECS", "Elitegroup Computer Sys" } ,
    { 0x101A, "NCR", "NCR/AT&T GIS" } ,
    { 0x101B, "Vitesse", "Vitesse Semiconductor" } ,
    { 0x101C, "WD", "Western Digital" } ,
    { 0x101E, "AMI", "American Megatrends Inc." } ,
    { 0x101F, "PictureTel", "PictureTel Corp." } ,
    { 0x1020, "Hitachi", "Hitachi Computer Electronics" } ,
    { 0x1021, "OKI", "Oki Electric Industry" } ,
    { 0x1022, "AMD", "Advanced Micro Devices" } ,
    { 0x1023, "Trident", "Trident Microsystems" } ,
    { 0x1024, "Zenith", "Zenith Data Systems" } ,
    { 0x1025, "Acer", "Acer Incorporated" } ,
    { 0x1028, "Dell", "Dell Computer Corporation" } ,
    { 0x1029, "Siem-Nix", "Siemens Nixdorf AG" } ,
    { 0x102A, "LSI", "LSI Logic Headland Div" } ,
    { 0x102B, "Matrox", "Matrox" } ,
    { 0x102C, "C&T", "Chips And Technologies" } ,
    { 0x102D, "Wyse", "Wyse Technologies" } ,
    { 0x102E, "Olivetti", "Olivetti Advanced Technology" } ,
    { 0x102F, "Toshiba", "Toshiba America" } ,
    { 0x1030, "TMC", "TMC Research" } ,
    { 0x1031, "miro", "miro Computer Products AG" } ,
    { 0x1032, "Compaq", "Compaq" } ,
    { 0x1033, "NEC", "NEC Corporation" } ,
    { 0x1034, "Burndy", "Burndy Corporation" } ,
    { 0x1035, "C&CRL", "Computer&Communication Research Lab" } ,
    { 0x1036, "FDomain", "Future Domain" } ,
    { 0x1037, "Hitachi", "Hitachi Micro Systems Inc" } ,
    { 0x1038, "AMP", "AMP Incorporated" } ,
    { 0x1039, "SiS", "Silicon Integrated System" } ,
    { 0x103A, "Seiko", "Seiko Epson Corporation" } ,
    { 0x103B, "Tatung", "Tatung Corp. Of America" } ,
    { 0x103C, "HP", "Hewlett-Packard Company" } ,
    { 0x103E, "Solliday", "Solliday Engineering" } ,
    { 0x103F, "Logic Mod.", "Logic Modeling" } ,
    { 0x1040, "Kubota", "Kubota Pacific Computer Inc." } ,
    { 0x1041, "Computrend", "Computrend" } ,
    { 0x1042, "PC Tech.", "PC Technology" } ,
    { 0x1043, "Asustek", "Asustek Computer Inc." } ,
    { 0x1044, "DPT", "Distributed Processing Tech" } ,
    { 0x1045, "OPTi", "OPTi Inc." } ,
    { 0x1046, "IPC", "IPC Corporation LTD" } ,
    { 0x1047, "Genoa", "Genoa Systems Corp." } ,
    { 0x1048, "ELSA", "ELSA AG" } ,
    { 0x1049, "Fountain", "Fountain Technology" } ,
    { 0x104A, "STM", "STMicroelectronics" } ,
    { 0x104B, "Mylex", "Mylex Corporation" } ,
    { 0x104C, "TI", "Texas Instruments" } ,
    { 0x104D, "Sony", "Sony Corporation" } ,
    { 0x104E, "Oak", "Oak Technology" } ,
    { 0x104F, "Co-Time", "Co-Time Computer Ltd." } ,
    { 0x1050, "Winbond", "Winbond Electronics Corp." } ,
    { 0x1051, "Anigma", "Anigma Corp." } ,
    { 0x1052, "Young", "Young Micro Systems" } ,
    { 0x1054, "Hitachi", "Hitachi LTD" } ,
    { 0x1055, "SMSC", "Standard Microsystems Corp." } ,
    { 0x1056, "ICL", "ICL" } ,
    { 0x1057, "Motorola", "Motorola" } ,
    { 0x1058, "E&TR", "Electronics & Telecommunication Res" } ,
    { 0x1059, "Teknor", "Teknor Microsystems" } ,
    { 0x105A, "Promise", "Promise Technology" } ,
    { 0x105B, "Foxconn", "Foxconn International" } ,
    { 0x105C, "Wipro", "Wipro Infotech Limited" } ,
    { 0x105D, "Number-Nine", "Number Nine Visual Technology" } ,
    { 0x105E, "Vtech", "Vtech Engineering Canada Ltd." } ,
    { 0x105F, "Infotronic", "Infotronic America Inc." } ,
    { 0x1060, "UMC", "United Microelectronics" } ,
    { 0x1061, "8x8", "8x8 Inc." } ,
    { 0x1062, "Maspar", "Maspar Computer Corp." } ,
    { 0x1063, "OOA", "Ocean Office Automation" } ,
    { 0x1064, "Alcatel", "Alcatel Cit" } ,
    { 0x1065, "TM", "Texas Microsystems" } ,
    { 0x1066, "Picopower", "Picopower Technology" } ,
    { 0x1067, "Mitsubishi", "Mitsubishi Electronics" } ,
    { 0x1068, "Div. Tech.", "Diversified Technology" } ,
    { 0x1069, "Mylex", "IBM Corporation" } ,
    { 0x106A, "Aten", "Aten Research Inc." } ,
    { 0x106B, "Apple", "Apple Computer Inc." } ,
    { 0x106C, "Hyundai", "Hyundai Electronics America" } ,
    { 0x106D, "Sequent", "Sequent" } ,
    { 0x106E, "DFI", "DFI Inc." } ,
    { 0x106F, "CityGate", "City Gate Development LTD" } ,
    { 0x1070, "Daewoo", "Daewoo Telecom Ltd." } ,
    { 0x1071, "Mitac", "Mitac" } ,
    { 0x1072, "GIT", "GIT Co. Ltd." } ,
    { 0x1073, "Yamaha", "Yamaha Corporation" } ,
    { 0x1074, "Nexgen", "Nexgen Microsysteme" } ,
    { 0x1075, "AIR", "Advanced Integration Research" } ,
    { 0x1076, "Chaintech", "Chaintech Computer Co. Ltd." } ,
    { 0x1077, "Q Logic", "Q Logic" } ,
    { 0x1078, "Cyrix", "Cyrix Corporation" } ,
    { 0x1079, "I-Bus", "I-Bus" } ,
    { 0x107A, "Networth", "Networth" } ,
    { 0x107B, "Gateway", "Gateway 2000" } ,
    { 0x107C, "Goldstar", "Goldstar Co. Ltd." } ,
    { 0x107D, "Leadtek", "Leadtek Research" } ,
    { 0x107E, "Interphase", "Interphase Corporation" } ,
    { 0x107F, "DTC", "Data Technology Corporation" } ,
    { 0x1080, "Contaq", "Contaq Microsystems" } ,
    { 0x1081, "Supermac", "Supermac Technology Inc." } ,
    { 0x1082, "EFA", "EFA Corporation Of America" } ,
    { 0x1083, "Forex", "Forex Computer Corporation" } ,
    { 0x1084, "Parador", "Parador" } ,
    { 0x1085, "Tulip", "Tulip Computers Int'l BV" } ,
    { 0x1086, "J. Bond", "J. Bond Computer Systems" } ,
    { 0x1087, "Cache", "Cache Computer" } ,
    { 0x1088, "MS Son", "Microcomputer Systems (M) Son" } ,
    { 0x1089, "DG", "Data General Corporation" } ,
    { 0x108A, "Bit3", "Bit3 Computer" } ,
    { 0x108C, "Elonex", "Elonex PLC c/o Oakleigh Systems Inc." } ,
    { 0x108D, "Olicom", "Olicom" } ,
    { 0x108E, "Sun", "Sun Microsystems" } ,
    { 0x108F, "Systemsoft", "Systemsoft Corporation" } ,
    { 0x1090, "Encore", "Encore Computer Corporation" } ,
    { 0x1091, "Intergraph", "Intergraph Corporation" } ,
    { 0x1092, "Diamond", "Diamond Computer Systems" } ,
    { 0x1093, "Nat. Inst.", "National Instruments" } ,
    { 0x1094, "FIC", "First Int'l Computers" } ,
    { 0x1095, "CMD", "CMD Technology Inc." } ,
    { 0x1096, "Alacron", "Alacron" } ,
    { 0x1097, "Appian", "Appian Graphics" } ,
    { 0x1098, "Quantum", "Quantum Designs Ltd." } ,
    { 0x1099, "Samsung", "Samsung Electronics Co. Ltd." } ,
    { 0x109A, "Packard-Bell", "Packard Bell" } ,
    { 0x109B, "Gemlight", "Gemlight Computer Ltd." } ,
    { 0x109C, "Megachips", "Megachips Corporation" } ,
    { 0x109D, "Zida", "Zida Technologies Ltd." } ,
    { 0x109E, "Brooktree", "Brooktree Corporation" } ,
    { 0x109F, "Trigem", "Trigem Computer Inc." } ,
    { 0x10A0, "Meidensha", "Meidensha Corporation" } ,
    { 0x10A1, "Juko", "Juko Electronics Inc. Ltd." } ,
    { 0x10A2, "Quantum", "Quantum Corporation" } ,
    { 0x10A3, "Everex", "Everex Systems Inc." } ,
    { 0x10A4, "Globe", "Globe Manufacturing Sales" } ,
    { 0x10A5, "Racal", "Racal Interlan" } ,
    { 0x10A6, "Informtech", "Informtech Industrial Ltd." } ,
    { 0x10A7, "Benchmarq", "Benchmarq Microelectronics" } ,
    { 0x10A8, "Sierra", "Sierra Semiconductor" } ,
    { 0x10A9, "SG", "Silicon Graphics" } ,
    { 0x10AA, "ACC", "ACC Microelectronics" } ,
    { 0x10AB, "Digicom", "Digicom" } ,
    { 0x10AC, "Honeywell", "Honeywell IASD" } ,
    { 0x10AD, "Symphony", "Symphony Labs" } ,
    { 0x10AE, "Cornerstone", "Cornerstone Technology" } ,
    { 0x10AF, "MCS", "Micro Computer Systems Inc." } ,
    { 0x10B0, "Cardexpert", "Cardexpert Technology" } ,
    { 0x10B1, "Cabletron", "Cabletron Systems Inc." } ,
    { 0x10B2, "Raytheon", "Raytheon Company" } ,
    { 0x10B3, "Databook", "Databook Inc." } ,
    { 0x10B4, "STB", "STB Systems" } ,
    { 0x10B5, "PLX", "PLX Technology" } ,
    { 0x10B6, "Madge", "Madge Networks" } ,
    { 0x10B7, "3Com", "3Com Corporation" } ,
    { 0x10B8, "SMC", "Standard Microsystems Corporation" } ,
    { 0x10B9, "ALi", "Acer Labs Inc." } ,
    { 0x10BA, "Mitsubishi", "Mitsubishi Electronics Corp." } ,
    { 0x10BB, "Dapha", "Dapha Electronics Corporation" } ,
    { 0x10BC, "ALR", "Advanced Logic Research Inc." } ,
    { 0x10BD, "Surecom", "Surecom Technology" } ,
    { 0x10BE, "Tseng", "Tsenglabs International Corp." } ,
    { 0x10BF, "MOST", "MOST Corp." } ,
    { 0x10C0, "Boca", "Boca Research Inc." } ,
    { 0x10C1, "ICM", "ICM Corp. Ltd." } ,
    { 0x10C2, "Auspex", "Auspex Systems Inc." } ,
    { 0x10C3, "Samsung", "Samsung Semiconductors" } ,
    { 0x10C4, "Award", "Award Software Int'l Inc." } ,
    { 0x10C5, "Xerox", "Xerox Corporation" } ,
    { 0x10C6, "Rambus", "Rambus Inc." } ,
    { 0x10C7, "Media Vision", "Media Vision" } ,
    { 0x10C8, "Neomagic", "Neomagic Corporation" } ,
    { 0x10C9, "Dataexpert", "Dataexpert Corporation" } ,
    { 0x10CA, "Fujitsu", "Fujitsu" } ,
    { 0x10CB, "Omron", "Omron Corporation" } ,
    { 0x10CC, "Mentor", "Mentor Arc Inc." } ,
    { 0x10CD, "AdvanSys", "Advanced System Products" } ,
    { 0x10CE, "Radius", "Radius Inc." } ,
    { 0x10CF, "Fujitsu", "Fujitsu Ltd." } ,
    { 0x10D0, "Fujitsu", "Fujitsu Limited" } ,
    { 0x10D1, "Future+", "Future+ Systems" } ,
    { 0x10D2, "Molex", "Molex Incorporated" } ,
    { 0x10D3, "Jabil", "Jabil Circuit Inc." } ,
    { 0x10D4, "Hualon", "Hualon Microelectronics" } ,
    { 0x10D5, "Autologic", "Autologic Inc." } ,
    { 0x10D6, "Cetia", "Cetia" } ,
    { 0x10D7, "BCM", "BCM Advanced Research" } ,
    { 0x10D8, "APL", "Advanced Peripherals Labs" } ,
    { 0x10D9, "Macronix", "Macronix International Co. Ltd." } ,
    { 0x10DA, "T-C", "Thomas-Conrad Corporation" } ,
    { 0x10DB, "Rohm", "Rohm Research" } ,
    { 0x10DC, "CERN", "CERN-European Lab. for Particle Physics" } ,
    { 0x10DD, "E&S", "Evans & Sutherland" } ,
    { 0x10DE, "Nvidia", "Nvidia Corporation" } ,
    { 0x10DF, "Emulex", "Emulex Corporation" } ,
    { 0x10E0, "IMS", "Integrated Micro Solutions" } ,
    { 0x10E1, "Tekram", "Tekram Technology Corp. Ltd." } ,
    { 0x10E2, "Aptix", "Aptix Corporation" } ,
    { 0x10E3, "Tundra", "Tundra Semiconductor Corp." } ,
    { 0x10E4, "Tandem", "Tandem Computers" } ,
    { 0x10E5, "MIC", "Micro Industries Corporation" } ,
    { 0x10E6, "Gainbery", "Gainbery Computer Products Inc." } ,
    { 0x10E7, "Vadem", "Vadem" } ,
    { 0x10E8, "AMCC", "Applied Micro Circuits Corp." } ,
    { 0x10E9, "Alps", "Alps Electronic Corp. Ltd." } ,
    { 0x10EA, "IGS", "InterGraphics Systems" } ,
    { 0x10EB, "Artist", "Artist Graphics" } ,
    { 0x10EC, "Realtek", "Realtek Semiconductor" } ,
    { 0x10ED, "Ascii", "Ascii Corporation" } ,
    { 0x10EE, "Xilinx", "Xilinx Corporation" } ,
    { 0x10EF, "Racore", "Racore Computer Products" } ,
    { 0x10F0, "Peritek", "Peritek Corporation" } ,
    { 0x10F1, "Tyan", "Tyan Computer" } ,
    { 0x10F2, "Achme", "Achme Computer Inc." } ,
    { 0x10F3, "Alaris", "Alaris Inc." } ,
    { 0x10F4, "S-Mos", "S-Mos Systems" } ,
    { 0x10F5, "NKK", "NKK Corporation" } ,
    { 0x10F6, "Creative", "Creative Electronic Systems SA" } ,
    { 0x10F7, "Matsushita", "Matsushita Electric Industrial Corp." } ,
    { 0x10F8, "Altos", "Altos India Ltd." } ,
    { 0x10F9, "PC-Direct", "PC Direct" } ,
    { 0x10FA, "Truevision", "Truevision" } ,
    { 0x10FB, "Thesys", "Thesys Microelectronic's" } ,
    { 0x10FC, "I-O", "I-O Data Device Inc." } ,
    { 0x10FD, "Soyo", "Soyo Technology Corp. Ltd." } ,
    { 0x10FE, "Fast", "Fast Electronic GmbH" } ,
    { 0x10FF, "Ncube", "Ncube" } ,
    { 0x1100, "Jazz", "Jazz Multimedia" } ,
    { 0x1101, "Initio", "Initio Corporation" } ,
    { 0x1102, "Creative Labs", "Creative Labs" } ,
    { 0x1103, "HighPoint", "HighPoint Technologies Inc." } ,
    { 0x1104, "Rasterops", "Rasterops" } ,
    { 0x1105, "Sigma", "Sigma Designs Inc." } ,
    { 0x1106, "VIA", "VIA Technologies Inc" } ,
    { 0x1107, "Stratus", "Stratus Computer" } ,
    { 0x1108, "Proteon", "Proteon Inc." } ,
    { 0x1109, "Cogent", "Cogent Data Technologies" } ,
    { 0x110A, "Infineon", "Infineon Technologies" } ,
    { 0x110B, "Chromatic", "Chromatic Research Inc" } ,
    { 0x110C, "Mini-Max", "Mini-Max Technology Inc." } ,
    { 0x110D, "ZNYX", "ZNYX Corporation" } ,
    { 0x110E, "CPU Tech.", "CPU Technology" } ,
    { 0x110F, "Ross", "Ross Technology" } ,
    { 0x1110, "Powerhouse", "Powerhouse Systems" } ,
    { 0x1111, "SCO", "Santa Cruz Operation" } ,
    { 0x1112, "Osicom", "Osicom Technologies Inc." } ,
    { 0x1113, "Accton", "Accton Technology Corporation" } ,
    { 0x1114, "Atmel", "Atmel Corp." } ,
    { 0x1115, "Dupont", "Dupont Pixel Systems Ltd." } ,
    { 0x1116, "Data Trans.", "Data Translation" } ,
    { 0x1117, "Datacube", "Datacube Inc." } ,
    { 0x1118, "Berg", "Berg Electronics" } ,
    { 0x1119, "Vortex", "ICP vortex Computersysteme GmbH" } ,
    { 0x111A, "Eff. Net.", "Efficent Networks" } ,
    { 0x111B, "Teledyne", "Teledyne Electronic Systems" } ,
    { 0x111C, "Tricord", "Tricord Systems Inc." } ,
    { 0x111D, "IDT", "Integrated Device Technology Inc." } ,
    { 0x111E, "Eldec", "Eldec Corp." } ,
    { 0x111F, "PDI", "Precision Digital Images" } ,
    { 0x1120, "EMC", "EMC Corp." } ,
    { 0x1121, "Zilog", "Zilog" } ,
    { 0x1122, "Multi-Tech", "Multi-Tech Systems Inc." } ,
    { 0x1123, "EDI", "Excellent Design Inc." } ,
    { 0x1124, "Leutron", "Leutron Vision AG" } ,
    { 0x1125, "Eurocore", "Eurocore/Vigra" } ,
    { 0x1126, "Vigra", "Vigra" } ,
    { 0x1127, "FORE", "FORE Systems" } ,
    { 0x1129, "Firmworks", "Firmworks" } ,
    { 0x112A, "Hermes", "Hermes Electronics Co. Ltd." } ,
    { 0x112B, "Linotype", "Linotype - Hell AG" } ,
    { 0x112C, "Zenith", "Zenith Data Systems" } ,
    { 0x112D, "Ravicad", "Ravicad" } ,
    { 0x112E, "Infomedia", "Infomedia" } ,
    { 0x112F, "ImagTech", "Imaging Technology" } ,
    { 0x1130, "Computervision", "Computervision" } ,
    { 0x1131, "Philips", "Philips Semiconductors" } ,
    { 0x1132, "Mitel", "Mitel Corp." } ,
    { 0x1133, "EIC", "Eicon Technology Corporation" } ,
    { 0x1134, "MCS", "Mercury Computer Systems Inc." } ,
    { 0x1135, "Fuji", "Fuji Xerox Co Ltd" } ,
    { 0x1136, "Momentum", "Momentum Data Systems" } ,
    { 0x1137, "Cisco", "Cisco Systems Inc" } ,
    { 0x1138, "Ziatech", "Ziatech Corporation" } ,
    { 0x1139, "Dyn. Pict.", "Dynamic Pictures Inc" } ,
    { 0x113A, "FWB", "FWB  Inc" } ,
    { 0x113B, "NCD", "Network Computing Devices" } ,
    { 0x113C, "Cyclone", "Cyclone Microsystems Inc." } ,
    { 0x113D, "Leading Edge", "Leading Edge Products Inc" } ,
    { 0x113E, "Sanyo", "Sanyo Electric Co" } ,
    { 0x113F, "Equinox", "Equinox Systems" } ,
    { 0x1140, "Intervoice", "Intervoice Inc" } ,
    { 0x1141, "Crest", "Crest Microsystem Inc" } ,
    { 0x1142, "Alliance", "Alliance Semiconductor CA - USA" } ,
    { 0x1143, "Netpower", "Netpower Inc" } ,
    { 0x1144, "Cinn. Mil.", "Cincinnati Milacron" } ,
    { 0x1145, "Workbit", "Workbit Corp" } ,
    { 0x1146, "Force", "Force Computers" } ,
    { 0x1147, "Interface", "Interface Corp" } ,
    { 0x1148, "S&K", "Schneider & Koch" } ,
    { 0x1149, "Win System", "Win System Corporation" } ,
    { 0x114A, "VMIC", "VMIC" } ,
    { 0x114B, "Canopus", "Canopus Co. Ltd" } ,
    { 0x114C, "Annabooks", "Annabooks" } ,
    { 0x114D, "IC Corp.", "IC Corporation" } ,
    { 0x114E, "Nikon", "Nikon Systems Inc" } ,
    { 0x114F, "Digi", "Digi International" } ,
    { 0x1150, "TMC", "Thinking Machines Corporation" } ,
    { 0x1151, "JAE", "JAE Electronics Inc." } ,
    { 0x1152, "Megatek", "Megatek" } ,
    { 0x1153, "Land Win", "Land Win Electronic Corp" } ,
    { 0x1154, "Melco", "Melco Inc" } ,
    { 0x1155, "Pine", "Pine Technology Ltd" } ,
    { 0x1156, "Periscope", "Periscope Engineering" } ,
    { 0x1157, "Avsys", "Avsys Corporation" } ,
    { 0x1158, "Voarx", "Voarx R&D Inc" } ,
    { 0x1159, "Mutech", "Mutech" } ,
    { 0x115A, "Harlequin", "Harlequin Ltd" } ,
    { 0x115B, "Parallax", "Parallax Graphics" } ,
    { 0x115C, "Photron", "Photron Ltd." } ,
    { 0x115D, "Xircom", "Xircom" } ,
    { 0x115E, "Peer", "Peer Protocols Inc" } ,
    { 0x115F, "Maxtor", "Maxtor Corporation" } ,
    { 0x1160, "Megasoft", "Megasoft Inc" } ,
    { 0x1161, "PFU", "PFU Ltd" } ,
    { 0x1162, "OA Lab", "OA Laboratory Co Ltd" } ,
    { 0x1163, "Rendition", "Rendition Inc" } ,
    { 0x1164, "APT", "Advanced Peripherals Tech" } ,
    { 0x1165, "Imagraph", "Imagraph Corporation" } ,
    { 0x1166, "RCC/ServerWorks", "Reliance Computer Corp./ServerWorks" } ,
    { 0x1167, "Mutoh", "Mutoh Industries Inc" } ,
    { 0x1168, "Thine", "Thine Electronics Inc" } ,
    { 0x1169, "CDAC", "Centre f/Dev. of Adv. Computing" } ,
    { 0x116A, "Polaris", "Polaris Communications" } ,
    { 0x116B, "Connectware", "Connectware Inc" } ,
    { 0x116C, "Int Res.", "Intelligent Resources" } ,
    { 0x116E, "EFI", "Electronics for Imaging" } ,
    { 0x116F, "WkSta. Tech.", "Workstation Technology" } ,
    { 0x1170, "Inventec", "Inventec Corporation" } ,
    { 0x1171, "Lough. Sound", "Loughborough Sound Images" } ,
    { 0x1172, "Altera", "Altera Corporation" } ,
    { 0x1173, "Adobe", "Adobe Systems" } ,
    { 0x1174, "Bridgeport", "Bridgeport Machines" } ,
    { 0x1175, "Mitron", "Mitron Computer Inc." } ,
    { 0x1176, "SBE", "SBE" } ,
    { 0x1177, "Silicon Eng.", "Silicon Engineering" } ,
    { 0x1178, "Alfa", "Alfa Inc" } ,
    { 0x1179, "Toshiba", "Toshiba America Info Systems" } ,
    { 0x117A, "A-Trend", "A-Trend Technology" } ,
    { 0x117B, "LG Elec.", "LG Electronics Inc." } ,
    { 0x117C, "Atto", "Atto Technology" } ,
    { 0x117D, "B&D", "Becton & Dickinson" } ,
    { 0x117E, "T/R", "T/R Systems" } ,
    { 0x117F, "ICS", "Integrated Circuit Systems" } ,
    { 0x1180, "CAC", "Communication Automation Corporation" } ,
    { 0x1181, "Telmatics", "Telmatics International" } ,
    { 0x1183, "Fujikura", "Fujikura Ltd" } ,
    { 0x1184, "Forks", "Forks Inc" } ,
    { 0x1185, "Dataworld", "Dataworld" } ,
    { 0x1186, "D-Link", "D-Link System Inc" } ,
    { 0x1187, "ATL", "Advanced Technology Laboratories" } ,
    { 0x1188, "Shima", "Shima Seiki Manufacturing Ltd." } ,
    { 0x1189, "Matsushita", "Matsushita Electronics" } ,
    { 0x118A, "Hilevel", "Hilevel Technology" } ,
    { 0x118B, "Hypertec", "Hypertec Pty Ltd" } ,
    { 0x118C, "Corollary", "Corollary Inc" } ,
    { 0x118D, "BitFlow", "BitFlow Inc" } ,
    { 0x118E, "Hermstedt", "Hermstedt GmbH" } ,
    { 0x118F, "Green", "Green Logic" } ,
    { 0x1190, "Tripace", "Tripace" } ,
    { 0x1191, "ACARD", "ACARD Technology" } ,
    { 0x1192, "Densan", "Densan Co. Ltd" } ,
    { 0x1193, "Zeitnet", "Zeitnet Inc." } ,
    { 0x1194, "Toucan", "Toucan Technology" } ,
    { 0x1195, "Ratoc", "Ratoc System Inc" } ,
    { 0x1196, "Hytec", "Hytec Electronics Ltd" } ,
    { 0x1197, "Gage", "Gage Applied Sciences Inc." } ,
    { 0x1198, "Lambda", "Lambda Systems Inc" } ,
    { 0x1199, "Attachmate", "Attachmate Corp." } ,
    { 0x119A, "Mind Share", "Mind Share Inc." } ,
    { 0x119B, "Omega", "Omega Micro Inc." } ,
    { 0x119C, "ITI", "Information Technology Inst." } ,
    { 0x119D, "Bug", "Bug Sapporo Japan" } ,
    { 0x119E, "Fujitsu", "Fujitsu Microelectronics Ltd." } ,
    { 0x119F, "Bull", "Bull Hn Information Systems" } ,
    { 0x11A0, "Convex", "Convex Computer Corporation" } ,
    { 0x11A1, "Hamamatsu", "Hamamatsu Photonics K.K." } ,
    { 0x11A2, "Sierra", "Sierra Research and Technology" } ,
    { 0x11A3, "Deuretzbacher", "Deuretzbacher GmbH & Co. Eng. KG" } ,
    { 0x11A4, "Barco", "Barco" } ,
    { 0x11A5, "MicroUnity", "MicroUnity Systems Engineering Inc." } ,
    { 0x11A6, "Pure Data", "Pure Data" } ,
    { 0x11A7, "Power Comp.", "Power Computing Corp." } ,
    { 0x11A8, "Systech", "Systech Corp." } ,
    { 0x11A9, "InnoSys", "InnoSys Inc." } ,
    { 0x11AA, "Actel", "Actel" } ,
    { 0x11AB, "Galileo", "Galileo Technology Ltd." } ,
    { 0x11AC, "Canon", "Canon Information Systems" } ,
    { 0x11AD, "Lite-On", "Lite-On Communications Inc" } ,
    { 0x11AE, "Scitex", "Scitex Corporation Ltd" } ,
    { 0x11AF, "Avid", "Avid Technology Inc." } ,
    { 0x11B0, "V3", "V3 Semiconductor Inc." } ,
    { 0x11B1, "Apricot", "Apricot Computers" } ,
    { 0x11B2, "Kodak", "Eastman Kodak" } ,
    { 0x11B3, "Barr", "Barr Systems Inc." } ,
    { 0x11B4, "Leitch", "Leitch Technology International" } ,
    { 0x11B5, "Radstone", "Radstone Technology Plc" } ,
    { 0x11B6, "United Video", "United Video Corp" } ,
    { 0x11B7, "Motorola", "Motorola" } ,
    { 0x11B8, "Xpoint", "Xpoint Technologies Inc" } ,
    { 0x11B9, "Pathlight", "Pathlight Technology Inc." } ,
    { 0x11BA, "Videotron", "Videotron Corp" } ,
    { 0x11BB, "Pyramid", "Pyramid Technology" } ,
    { 0x11BC, "Net. Periph.", "Network Peripherals Inc" } ,
    { 0x11BD, "Pinnacle", "Pinnacle Systems Inc." } ,
    { 0x11BE, "IMI", "International Microcircuits Inc" } ,
    { 0x11BF, "Astrodesign", "Astrodesign Inc." } ,
    { 0x11C0, "HP", "Hewlett-Packard" } ,
    { 0x11C1, "Agere", "Agere Systems" } ,
    { 0x11C2, "Sand", "Sand Microelectronics" } ,
    { 0x11C3, "NEC", "NEC Corporation" } ,
    { 0x11C4, "Doc. Tech.", "Document Technologies Ind." } ,
    { 0x11C5, "Shiva", "Shiva Corporatin" } ,
    { 0x11C6, "Dainippon", "Dainippon Screen Mfg. Co" } ,
    { 0x11C7, "D.C.M.", "D.C.M. Data Systems" } ,
    { 0x11C8, "Dolphin", "Dolphin Interconnect Solutions" } ,
    { 0x11C9, "MAGMA", "MAGMA" } ,
    { 0x11CA, "LSI Sys.", "LSI Systems Inc" } ,
    { 0x11CB, "Specialix", "Specialix International Ltd." } ,
    { 0x11CC, "M&K", "Michels & Kleberhoff Computer GmbH" } ,
    { 0x11CD, "HAL", "HAL Computer Systems Inc." } ,
    { 0x11CE, "PRI", "Primary Rate Inc" } ,
    { 0x11CF, "PEC", "Pioneer Electronic Corporation" } ,
    { 0x11D0, "Loral", "Loral Frederal Systems - Manassas" } ,
    { 0x11D1, "AuraVision", "AuraVision Corporation" } ,
    { 0x11D2, "Intercom", "Intercom Inc." } ,
    { 0x11D3, "Trancell", "Trancell Systems Inc" } ,
    { 0x11D4, "AD", "Analog Devices" } ,
    { 0x11D5, "Ikon", "Ikon Corp" } ,
    { 0x11D6, "Tekelec", "Tekelec Technologies" } ,
    { 0x11D7, "Trenton", "Trenton Terminals Inc" } ,
    { 0x11D8, "ITD", "Image Technologies Development" } ,
    { 0x11D9, "Tec", "Tec Corporation" } ,
    { 0x11DA, "Novell", "Novell" } ,
    { 0x11DB, "Sega", "Sega Enterprises Ltd" } ,
    { 0x11DC, "Questra", "Questra Corp" } ,
    { 0x11DD, "Crosfield", "Crosfield Electronics Ltd" } ,
    { 0x11DE, "Zoran", "Zoran Corporation" } ,
    { 0x11DF, "New Wave", "New Wave Pdg" } ,
    { 0x11E0, "Cray", "Cray Communications A/S" } ,
    { 0x11E1, "Gec Plessey", "Gec Plessey Semi Inc" } ,
    { 0x11E2, "Samsung", "Samsung Information Systems America" } ,
    { 0x11E3, "Quicklogic", "Quicklogic Corp" } ,
    { 0x11E4, "Second Wave", "Second Wave Inc" } ,
    { 0x11E5, "IIX", "IIX Consulting" } ,
    { 0x11E6, "Mitsui", "Mitsui-Zosen System Research" } ,
    { 0x11E7, "Toshiba", "Toshiba America Elec. Co" } ,
    { 0x11E8, "DPSI", "Digital Processing Systems Inc" } ,
    { 0x11E9, "Highwater", "Highwater Designs Ltd" } ,
    { 0x11EA, "Elsag", "Elsag Bailey" } ,
    { 0x11EB, "Formation", "Formation Inc" } ,
    { 0x11EC, "Coreco", "Coreco Inc" } ,
    { 0x11ED, "Mediamatics", "Mediamatics" } ,
    { 0x11EE, "Dome", "Dome Imaging Systems Inc" } ,
    { 0x11EF, "Nicolet", "Nicolet Technologies BV" } ,
    { 0x11F0, "Compu-Shack", "Compu-Shack GmbH" } ,
    { 0x11F1, "Symbios", "Symbios Logic Inc" } ,
    { 0x11F2, "Pic-Tel", "Picture Tel Japan KK" } ,
    { 0x11F3, "Keithley", "Keithley Metrabyte" } ,
    { 0x11F4, "Kinetic", "Kinetic Systems Corporation" } ,
    { 0x11F5, "Comp Dev", "Computing Devices Intl" } ,
    { 0x11F6, "Powermatic", "Powermatic Data Systems Ltd" } ,
    { 0x11F7, "S-A", "Scientific Atlanta" } ,
    { 0x11F8, "PMC-Sierra", "PMC-Sierra Inc." } ,
    { 0x11F9, "I-Cube", "I-Cube Inc" } ,
    { 0x11FA, "Kasan", "Kasan Electronics Co Ltd" } ,
    { 0x11FB, "Datel", "Datel Inc" } ,
    { 0x11FC, "Silicon Magic", "Silicon Magic" } ,
    { 0x11FD, "High Street", "High Street Consultants" } ,
    { 0x11FE, "Comtrol", "Comtrol Corp" } ,
    { 0x11FF, "Scion", "Scion Corp" } ,
    { 0x1200, "CSS", "CSS Corp" } ,
    { 0x1201, "Vista", "Vista Controls Corp" } ,
    { 0x1202, "Network Gen", "Network General Corp" } ,
    { 0x1203, "Agfa", "Bayer Corporation Agfa Div" } ,
    { 0x1204, "Lattice", "Lattice Semiconductor Corp" } ,
    { 0x1205, "Array", "Array Corp" } ,
    { 0x1206, "Amdahl", "Amdahl Corp" } ,
    { 0x1208, "Parsytec", "Parsytec GmbH" } ,
    { 0x1209, "Sci Sys", "Sci Systems Inc" } ,
    { 0x120A, "Synaptel", "Synaptel" } ,
    { 0x120B, "Adaptive", "Adaptive Solutions" } ,
    { 0x120D, "Comp Labs", "Compression Labs Inc." } ,
    { 0x120E, "Cyclades", "Cyclades Corporation" } ,
    { 0x120F, "Essential", "Essential Communications" } ,
    { 0x1210, "Hyperparallel", "Hyperparallel Technologies" } ,
    { 0x1211, "Braintech", "Braintech Inc" } ,
    { 0x1212, "Kingston", "Kingston Technology Corp" } ,
    { 0x1213, "AISI", "Applied Intelligent Systems Inc" } ,
    { 0x1214, "Perf Tech", "Performance Technologies Inc" } ,
    { 0x1215, "Interware", "Interware Co Ltd" } ,
    { 0x1216, "Purup Eskofot", "Purup-Eskofot A/S" } ,
    { 0x1217, "O2Micro", "O2Micro Inc" } ,
    { 0x1218, "Hybricon", "Hybricon Corp" } ,
    { 0x1219, "First Virtual", "First Virtual Corp" } ,
    { 0x121A, "3dfx", "3dfx Interactive Inc" } ,
    { 0x121B, "ATM", "Advanced Telecommunications Modules" } ,
    { 0x121C, "Nippon Texa", "Nippon Texa Co Ltd" } ,
    { 0x121D, "Lippert", "Lippert Automationstechnik GmbH" } ,
    { 0x121E, "CSPI", "CSPI" } ,
    { 0x121F, "Arcus", "Arcus Technology Inc" } ,
    { 0x1220, "Ariel", "Ariel Corporation" } ,
    { 0x1221, "Contec", "Contec Co Ltd" } ,
    { 0x1222, "Ancor", "Ancor Communications Inc" } ,
    { 0x1223, "Heurikon", "Heurikon/Computer Products" } ,
    { 0x1224, "Int. Img.", "Interactive Images" } ,
    { 0x1225, "Power IO", "Power I/O Inc." } ,
    { 0x1227, "Tech-Source", "Tech-Source" } ,
    { 0x1228, "Norsk", "Norsk Elektro Optikk A/S" } ,
    { 0x1229, "Data Kin", "Data Kinesis Inc." } ,
    { 0x122A, "Int. Telecom", "Integrated Telecom" } ,
    { 0x122B, "LG Ind.", "LG Industrial Systems Co. Ltd." } ,
    { 0x122C, "sci-worx", "sci-worx GmbH" } ,
    { 0x122D, "Aztech", "Aztech System Ltd" } ,
    { 0x122E, "Xyratex", "Xyratex" } ,
    { 0x122F, "Andrew", "Andrew Corp." } ,
    { 0x1230, "Fishcamp", "Fishcamp Engineering" } ,
    { 0x1231, "WMI", "Woodward McCoach Inc." } ,
    { 0x1232, "GPT", "GPT Ltd." } ,
    { 0x1233, "Bus-Tech", "Bus-Tech Inc." } ,
    { 0x1234, "Technical", "Technical Corp" } ,
    { 0x1235, "Risq Mod", "Risq Modular Systems Inc." } ,
    { 0x1236, "Sigma", "Sigma Designs Corp." } ,
    { 0x1237, "Alta Tech", "Alta Technology Corp." } ,
    { 0x1238, "Adtran", "Adtran" } ,
    { 0x1239, "3DO", "The 3DO Company" } ,
    { 0x123A, "Visicom", "Visicom Laboratories Inc." } ,
    { 0x123B, "Seeq", "Seeq Technology Inc." } ,
    { 0x123C, "Century Sys", "Century Systems Inc." } ,
    { 0x123D, "EDT", "Engineering Design Team Inc." } ,
    { 0x123F, "C-Cube", "C-Cube Microsystems" } ,
    { 0x1240, "Marathon", "Marathon Technologies Corp." } ,
    { 0x1241, "DSC", "DSC Communications" } ,
    { 0x1242, "JNI", "JNI Corporation" } ,
    { 0x1243, "Delphax", "Delphax" } ,
    { 0x1244, "AVM", "AVM AUDIOVISUELLES MKTG & Computer GmbH" } ,
    { 0x1245, "APD", "APD S.A." } ,
    { 0x1246, "Dipix", "Dipix Technologies Inc" } ,
    { 0x1247, "Xylon", "Xylon Research Inc." } ,
    { 0x1248, "Central Data", "Central Data Corp." } ,
    { 0x1249, "Samsung", "Samsung Electronics Co. Ltd." } ,
    { 0x124A, "AEG", "AEG Electrocom GmbH" } ,
    { 0x124B, "GreenSpring", "GreenSpring Computers" } ,
    { 0x124C, "Solitron", "Solitron Technologies Inc." } ,
    { 0x124D, "Stallion", "Stallion Technologies" } ,
    { 0x124E, "Cylink", "Cylink" } ,
    { 0x124F, "Infortrend", "Infortrend Technology Inc" } ,
    { 0x1250, "Hitachi", "Hitachi Microcomputer System Ltd." } ,
    { 0x1251, "VLSI Sol.", "VLSI Solution OY" } ,
    { 0x1253, "Guzik", "Guzik Technical Enterprises" } ,
    { 0x1254, "Linear Systems", "Linear Systems Ltd." } ,
    { 0x1255, "Optibase", "Optibase Ltd." } ,
    { 0x1256, "Perceptive", "Perceptive Solutions Inc." } ,
    { 0x1257, "Vertex", "Vertex Networks Inc." } ,
    { 0x1258, "Gilbarco", "Gilbarco Inc." } ,
    { 0x1259, "Allied Tsyn", "Allied Telesyn International" } ,
    { 0x125A, "ABB Pwr", "ABB Power Systems" } ,
    { 0x125B, "Asix", "Asix Electronics Corp." } ,
    { 0x125C, "Aurora", "Aurora Technologies Inc." } ,
    { 0x125D, "ESS", "ESS Technology" } ,
    { 0x125E, "Specvideo", "Specialvideo Engineering SRL" } ,
    { 0x125F, "Concurrent", "Concurrent Technologies Inc." } ,
    { 0x1260, "Intersil", "Intersil Corporation" } ,
    { 0x1261, "Matsushita", "Matsushita-Kotobuki Electronics Indu" } ,
    { 0x1262, "ES Comp.", "ES Computer Co. Ltd." } ,
    { 0x1263, "Sonic Sol.", "Sonic Solutions" } ,
    { 0x1264, "Aval Nag.", "Aval Nagasaki Corp." } ,
    { 0x1265, "Casio", "Casio Computer Co. Ltd." } ,
    { 0x1266, "Microdyne", "Microdyne Corp." } ,
    { 0x1267, "SA Telecom", "S.A. Telecommunications" } ,
    { 0x1268, "Tektronix", "Tektronix" } ,
    { 0x1269, "Thomson-CSF", "Thomson-CSF/TTM" } ,
    { 0x126A, "Lexmark", "Lexmark International Inc." } ,
    { 0x126B, "Adax", "Adax Inc." } ,
    { 0x126C, "NorTel", "Northern Telecom" } ,
    { 0x126D, "Splash", "Splash Technology Inc." } ,
    { 0x126E, "Sumitomo", "Sumitomo Metal Industries Ltd." } ,
    { 0x126F, "Sil Motion", "Silicon Motion" } ,
    { 0x1270, "Olympus", "Olympus Optical Co. Ltd." } ,
    { 0x1271, "GW Instr.", "GW Instruments" } ,
    { 0x1272, "Telematics", "Telematics International" } ,
    { 0x1273, "Hughes", "Hughes Network Systems" } ,
    { 0x1274, "Ensoniq", "Ensoniq" } ,
    { 0x1275, "NetApp", "Network Appliance" } ,
    { 0x1276, "Sw Net Tech", "Switched Network Technologies Inc." } ,
    { 0x1277, "Comstream", "Comstream" } ,
    { 0x1278, "Transtech", "Transtech Parallel Systems" } ,
    { 0x1279, "Transmeta", "Transmeta Corp." } ,
    { 0x127A, "Conexant", "Conexant Systems" } ,
    { 0x127B, "Pixera", "Pixera Corp" } ,
    { 0x127C, "Crosspoint", "Crosspoint Solutions Inc." } ,
    { 0x127D, "Vela Res", "Vela Research" } ,
    { 0x127E, "Winnow", "Winnov L.P." } ,
    { 0x127F, "Fujifilm", "Fujifilm" } ,
    { 0x1280, "Photoscript", "Photoscript Group Ltd." } ,
    { 0x1281, "Yokogawa", "Yokogawa Electronic Corp." } ,
    { 0x1282, "Davicom", "Davicom Semiconductor Inc." } ,
    { 0x1283, "ITExpress", "Integrated Technology Express Inc." } ,
    { 0x1284, "Sahara", "Sahara Networks Inc." } ,
    { 0x1285, "Plat Tech", "Platform Technologies Inc." } ,
    { 0x1286, "Mazet", "Mazet GmbH" } ,
    { 0x1287, "LuxSonor", "LuxSonor Inc." } ,
    { 0x1288, "Timestep", "Timestep Corp." } ,
    { 0x1289, "AVC Tech", "AVC Technology Inc." } ,
    { 0x128A, "Asante", "Asante Technologies Inc." } ,
    { 0x128B, "Transwitch", "Transwitch Corp." } ,
    { 0x128C, "Retix", "Retix Corp." } ,
    { 0x128D, "G2 Net", "G2 Networks Inc." } ,
    { 0x128E, "Samho", "Samho Multi Tech Ltd." } ,
    { 0x128F, "Tateno", "Tateno Dennou Inc." } ,
    { 0x1290, "Sord", "Sord Computer Corp." } ,
    { 0x1291, "NCS Comp", "NCS Computer Italia" } ,
    { 0x1292, "Tritech", "Tritech Microelectronics Intl PTE" } ,
    { 0x1293, "M Reality", "Media Reality Technology" } ,
    { 0x1294, "Rhetorex", "Rhetorex Inc." } ,
    { 0x1295, "Imagenation", "Imagenation Corp." } ,
    { 0x1296, "Kofax", "Kofax Image Products" } ,
    { 0x1297, "Holco", "Holco Enterprise" } ,
    { 0x1298, "Spellcaster", "Spellcaster Telecommunications Inc." } ,
    { 0x1299, "Know Tech", "Knowledge Technology Laboratories" } ,
    { 0x129A, "VMETRO", "VMETRO Inc." } ,
    { 0x129B, "Img Access", "Image Access" } ,
    { 0x129D, "CompCore", "CompCore Multimedia Inc." } ,
    { 0x129E, "Victor Jpn", "Victor Co. of Japan Ltd." } ,
    { 0x129F, "OEC Med", "OEC Medical Systems Inc." } ,
    { 0x12A0, "A-B", "Allen Bradley Co." } ,
    { 0x12A1, "Simpact", "Simpact Inc" } ,
    { 0x12A2, "NewGen", "NewGen Systems Corp." } ,
    { 0x12A3, "Lucent", "Lucent Technologies AMR" } ,
    { 0x12A4, "NTT Elect", "NTT Electronics Technology Co." } ,
    { 0x12A5, "Vision Dyn", "Vision Dynamics Ltd." } ,
    { 0x12A6, "Scalable", "Scalable Networks Inc." } ,
    { 0x12A7, "AMO", "AMO GmbH" } ,
    { 0x12A8, "News Datacom", "News Datacom" } ,
    { 0x12A9, "Xiotech", "Xiotech Corp." } ,
    { 0x12AA, "SDL", "SDL Communications Inc." } ,
    { 0x12AB, "Yuan Yuan", "Yuan Yuan Enterprise Co. Ltd." } ,
    { 0x12AC, "MeasureX", "MeasureX Corp." } ,
    { 0x12AD, "Multidata", "Multidata GmbH" } ,
    { 0x12AE, "Alteon", "Alteon Networks Inc." } ,
    { 0x12AF, "TDK USA", "TDK USA Corp." } ,
    { 0x12B0, "Jorge Sci", "Jorge Scientific Corp." } ,
    { 0x12B1, "GammaLink", "GammaLink" } ,
    { 0x12B2, "Gen Signal", "General Signal Networks" } ,
    { 0x12B3, "Inter-Face", "Inter-Face Co. Ltd." } ,
    { 0x12B4, "Future Tel", "Future Tel Inc." } ,
    { 0x12B5, "Granite", "Granite Systems Inc." } ,
    { 0x12B6, "Nat Micro", "Natural Microsystems" } ,
    { 0x12B7, "Acumen", "Acumen" } ,
    { 0x12B8, "Korg", "Korg" } ,
    { 0x12B9, "US Robotics", "US Robotics" } ,
    { 0x12BA, "Bittware", "Bittware Research Systems Inc" } ,
    { 0x12BB, "Nippon Uni", "Nippon Unisoft Corp." } ,
    { 0x12BC, "Array Micro", "Array Microsystems" } ,
    { 0x12BD, "Computerm", "Computerm Corp." } ,
    { 0x12BE, "Anchor Chips", "Anchor Chips Inc." } ,
    { 0x12BF, "Fujifilm", "Fujifilm Microdevices" } ,
    { 0x12C0, "Infimed", "Infimed" } ,
    { 0x12C1, "GMM Res", "GMM Research Corp." } ,
    { 0x12C2, "Mentec", "Mentec Ltd." } ,
    { 0x12C3, "Holtek", "Holtek Microelectronics Inc." } ,
    { 0x12C4, "Conn Tech", "Connect Tech Inc." } ,
    { 0x12C5, "PicturEl", "Picture Elements Inc." } ,
    { 0x12C6, "Mitani", "Mitani Corp." } ,
    { 0x12C7, "Dialogic", "Dialogic Corp." } ,
    { 0x12C8, "G Force", "G Force Co. Ltd." } ,
    { 0x12C9, "Gigi Ops", "Gigi Operations" } ,
    { 0x12CA, "ICE", "Integrated Computing Engines, Inc." } ,
    { 0x12CB, "Antex", "Antex Electronics Corp." } ,
    { 0x12CC, "Pluto", "Pluto Technologies International" } ,
    { 0x12CD, "Aims Lab", "Aims Lab" } ,
    { 0x12CE, "Netspeed", "Netspeed Inc." } ,
    { 0x12CF, "Prophet", "Prophet Systems Inc." } ,
    { 0x12D0, "GDE Sys", "GDE Systems Inc." } ,
    { 0x12D1, "PsiTech", "PsiTech" } ,
    { 0x12D2, "NVidia", "NVidia / SGS Thomson" } ,
    { 0x12D3, "Vingmed", "Vingmed Sound A/S" } ,
    { 0x12D4, "DGM&S", "DGM & S" } ,
    { 0x12D5, "Equator", "Equator Technologies" } ,
    { 0x12D6, "Analogic", "Analogic Corp." } ,
    { 0x12D7, "Biotronic", "Biotronic SRL" } ,
    { 0x12D8, "Pericom", "Pericom Semiconductor" } ,
    { 0x12D9, "Aculab", "Aculab Plc." } ,
    { 0x12DA, "TrueTime", "TrueTime" } ,
    { 0x12DB, "Annapolis", "Annapolis Micro Systems Inc." } ,
    { 0x12DC, "Symicron", "Symicron Computer Communication Ltd." } ,
    { 0x12DD, "MGI", "Management Graphics Inc." } ,
    { 0x12DE, "Rainbow", "Rainbow Technologies" } ,
    { 0x12DF, "SBS Tech", "SBS Technologies Inc." } ,
    { 0x12E0, "Chase", "Chase Research PLC" } ,
    { 0x12E1, "Nintendo", "Nintendo Co. Ltd." } ,
    { 0x12E2, "Datum", "Datum Inc. Bancomm-Timing Division" } ,
    { 0x12E3, "Imation", "Imation Corp. - Medical Imaging Syst" } ,
    { 0x12E4, "Brooktrout", "Brooktrout Technology Inc." } ,
    { 0x12E6, "Cirel", "Cirel Systems" } ,
    { 0x12E7, "Sebring", "Sebring Systems Inc" } ,
    { 0x12E8, "CRISC", "CRISC Corp." } ,
    { 0x12E9, "GE Spacenet", "GE Spacenet" } ,
    { 0x12EA, "Zuken", "Zuken" } ,
    { 0x12EB, "Aureal", "Aureal Semiconductor" } ,
    { 0x12EC, "3A Intl", "3A International Inc." } ,
    { 0x12ED, "Optivision", "Optivision Inc." } ,
    { 0x12EE, "Orange Micro", "Orange Micro, Inc." } ,
    { 0x12EF, "Vienna", "Vienna Systems" } ,
    { 0x12F0, "Pentek", "Pentek" } ,
    { 0x12F1, "Sorenson", "Sorenson Vision Inc." } ,
    { 0x12F2, "Gammagraphx", "Gammagraphx Inc." } ,
    { 0x12F4, "Megatel", "Megatel" } ,
    { 0x12F5, "Forks", "Forks" } ,
    { 0x12F6, "Dawson Fr", "Dawson France" } ,
    { 0x12F7, "Cognex", "Cognex" } ,
    { 0x12F8, "Electronic-Design", "Electronic-Design GmbH" } ,
    { 0x12F9, "FFT", "FourFold Technologies" } ,
    { 0x12FB, "SSP", "Spectrum Signal Processing" } ,
    { 0x12FC, NULL, "Capital Equipment Corp" } ,
    { 0x12FE, "ESD", "ESD Electronic System Design GmbH" } ,
    { 0x1304, NULL, "Juniper Networks Inc." } ,
    { 0x1307, "ComputerBoards", "ComputerBoards" } ,
    { 0x1308, "Jato", "Jato Technologies Inc." } ,
    { 0x130A, NULL, "Mitsubishi Electric Microcomputer" } ,
    { 0x130B, NULL, "Colorgraphic Communications Corp" } ,
    { 0x130F, NULL, "Advanet Inc." } ,
    { 0x1310, NULL, "Gespac" } ,
    { 0x1312, "RVSI", "Robotic Vision Systems Incorporated" } ,
    { 0x1313, NULL, "Yaskawa Electric Co." } ,
    { 0x1316, NULL, "Teradyne Inc." } ,
    { 0x1317, NULL, "Admtek Inc" } ,
    { 0x1318, "Packet Engines", "Packet Engines, Inc." } ,
    { 0x1319, "Forte Media", "Forte Media, Inc." } ,
    { 0x131f, NULL, "SIIG" } ,
    { 0x1325, NULL, "Salix Technologies Inc" } ,
    { 0x1326, NULL, "Seachange International" } ,
    { 0x1331, "RadiSys", "RadiSys Corporation" } ,
    { 0x1335, "Videomail", "Videomail Inc." } ,
    { 0x133D, NULL, "Prisa Networks" } ,
    { 0x133F, NULL, "SCM Microsystems" } ,
    { 0x1342, NULL, "Promax Systems Inc" } ,
    { 0x1344, "Micron", "Micron Technology, Inc." } ,
    { 0x1347, "Odetics", "Odetics" } ,
    { 0x134A, "DTC", "DTC Technology Corp." } ,
    { 0x134B, NULL, "ARK Research Corp." } ,
    { 0x134C, NULL, "Chori Joho System Co. Ltd" } ,
    { 0x134D, "PCTEL", "PCTEL Inc." } ,
    { 0x135A, NULL, "Brain Boxes Limited" } ,
    { 0x135B, NULL, "Giganet Inc." } ,
    { 0x135C, NULL, "Quatech Inc" } ,
    { 0x135D, "ABB Network Partn", "ABB Network Partner AB" } ,
    { 0x135E, "Sealevel", "Sealevel Systems Inc." } ,
    { 0x135F, NULL, "I-Data International A-S" } ,
    { 0x1360, NULL, "Meinberg Funkuhren" } ,
    { 0x1361, NULL, "Soliton Systems K.K." } ,
    { 0x1363, NULL, "Phoenix Technologies Ltd" } ,
    { 0x1365, "Hypercope", "Hypercope Corp." } ,
    { 0x1366, "Teijin", "Teijin Seiki Co. Ltd." } ,
    { 0x1367, NULL, "Hitachi Zosen Corporation" } ,
    { 0x1368, NULL, "Skyware Corporation" } ,
    { 0x1369, NULL, "Digigram" } ,
    { 0x136B, NULL, "Kawasaki Steel Corporation" } ,
    { 0x136C, NULL, "Adtek System Science Co Ltd" } ,
    { 0x1375, NULL, "Boeing - Sunnyvale" } ,
    { 0x1377, NULL, "GMBH" } ,
    { 0x137A, NULL, "Mark Of The Unicorn Inc" } ,
    { 0x137B, NULL, "PPT Vision" } ,
    { 0x137C, NULL, "Iwatsu Electric Co Ltd" } ,
    { 0x137D, NULL, "Dynachip Corporation" } ,
    { 0x137E, "PTSC", "Patriot Scientific Corp." } ,
    { 0x1380, NULL, "Sanritz Automation Co LTC" } ,
    { 0x1381, NULL, "Brains Co. Ltd" } ,
    { 0x1382, "Marian", "Marian - Electronic & Software" } ,
    { 0x1384, NULL, "Stellar Semiconductor Inc" } ,
    { 0x1385, "Netgear", "Netgear" } ,
    { 0x1387, NULL, "Systran Corp" } ,
    { 0x1388, NULL, "Hitachi Information Technology Co Ltd" } ,
    { 0x1389, "Applicom", "Applicom International" } ,
    { 0x138B, NULL, "Tokimec Inc" } ,
    { 0x138E, NULL, "Basler GMBH" } ,
    { 0x138F, NULL, "Patapsco Designs Inc" } ,
    { 0x1390, "CDI", "Concept Development Inc." } ,
    { 0x1393, NULL, "Moxa Technologies Co Ltd" } ,
    { 0x1395, NULL, "Ambicom Inc" } ,
    { 0x1396, NULL, "Cipher Systems Inc" } ,
    { 0x1397, "Cologne", "Cologne Chip Designs GmbH" } ,
    { 0x1398, NULL, "Clarion Co. Ltd" } ,
    { 0x139A, NULL, "Alacritech Inc" } ,
    { 0x139D, NULL, "Xstreams PLC/ EPL Limited" } ,
    { 0x139E, NULL, "Echostar Data Networks" } ,
    { 0x13A0, NULL, "Crystal Group Inc" } ,
    { 0x13A1, NULL, "Kawasaki Heavy Industries Ltd" } ,
    { 0x13A3, "HI-FN", "HI-FN Inc." } ,
    { 0x13A4, NULL, "Rascom Inc" } ,
    { 0x13A7, NULL, "Teles AG" } ,
    { 0x13A8, "XR", "Exar Corp." } ,
    { 0x13A9, NULL, "Siemens Medical Systems Ultrasound Group" } ,
    { 0x13AA, NULL, "Nortel Networks - BWA Division" } ,
    { 0x13AF, NULL, "T.Sqware" } ,
    { 0x13B1, NULL, "Tamura Corporation" } ,
    { 0x13B4, NULL, "Wellbean Co Inc" } ,
    { 0x13B5, NULL, "ARM Ltd" } ,
    { 0x13B6, NULL, "DLoG GMBH" } ,
    { 0x13B8, NULL, "Nokia Telecommunications OY" } ,
    { 0x13BD, "SHARP", "Sharp Corporation" } ,
    { 0x13BF, NULL, "Sharewave Inc" } ,
    { 0x13C0, "Microgate", "Microgate Corp." } ,
    { 0x13C1, "3ware", "3ware Inc." } ,
    { 0x13C2, NULL, "Technotrend Systemtechnik GMBH" } ,
    { 0x13C3, NULL, "Janz Computer AG" } ,
    { 0x13C7, NULL, "Blue Chip Technology Ltd" } ,
    { 0x13CC, NULL, "Metheus Corporation" } ,
    { 0x13CF, NULL, "Studio Audio & Video Ltd" } ,
    { 0x13D0, NULL, "B2C2 Inc" } ,
    { 0x13D1, NULL, "Abocom Systems Inc" } ,
    { 0x13D4, NULL, "Graphics Microsystems Inc" } ,
    { 0x13D6, NULL, "K.I. Technology Co Ltd" } ,
    { 0x13D7, NULL, "Toshiba Engineering Corporation" } ,
    { 0x13D8, NULL, "Phobos Corporation" } ,
    { 0x13D9, NULL, "Apex Inc" } ,
    { 0x13DC, NULL, "Netboost Corporation" } ,
    { 0x13DE, NULL, "ABB Robotics Products AB" } ,
    { 0x13DF, "E-Tech", "E-Tech Inc." } ,
    { 0x13E0, NULL, "GVC Corporation" } ,
    { 0x13E3, NULL, "Nest Inc" } ,
    { 0x13E4, NULL, "Calculex Inc" } ,
    { 0x13E5, NULL, "Telesoft Design Ltd" } ,
    { 0x13E9, NULL, "Intraserver Technology Inc" } ,
    { 0x13EA, NULL, "Dallas Semiconductor" } ,
    { 0x13F0, NULL, "Sundance Technology Inc" } ,
    { 0x13F1, NULL, "OCE - Industries S.A." } ,
    { 0x13F4, NULL, "Troika Networks Inc" } ,
    { 0x13F6, "C-Media", "C-Media Electronics Inc." } ,
    { 0x13F9, NULL, "NTT Advanced Technology Corp." } ,
    { 0x13FA, "Pentland", "Pentland Systems Ltd." } ,
    { 0x13FB, NULL, "Aydin Corp" } ,
    { 0x13FD, NULL, "Micro Science Inc" } ,
    { 0x13FE, "Advantech", "Advantech Co., Ltd." } ,
    { 0x1400, NULL, "ARTX Inc" } ,
    { 0x1402, "Meilhaus Electronic", "Meilhaus Electronic GmbH Germany" } ,
    { 0x1404, NULL, "Fundamental Software Inc" } ,
    { 0x1406, "Oc", "Oc Printing Systems" } ,
    { 0x1407, "LAVA", "Lava Computer MFG Inc." } ,
    { 0x1408, NULL, "Aloka Co. Ltd" } ,
    { 0x1409, NULL, "eTIMedia Technology Co Ltd" } ,
    { 0x140A, NULL, "DSP Research Inc" } ,
    { 0x140B, NULL, "Ramix Inc" } ,
    { 0x140D, NULL, "Matsushita Electric Works Ltd" } ,
    { 0x140F, NULL, "Salient Systems Corp" } ,
    { 0x1412, "IC Ensemble", "IC Ensemble, Inc." } ,
    { 0x1413, NULL, "Addonics" } ,
    { 0x1415, NULL, "Oxford Semiconductor Ltd" } ,
    { 0x1418, NULL, "Kyushu Electronics Systems Inc" } ,
    { 0x1419, NULL, "Excel Switching Corp" } ,
    { 0x141B, NULL, "Zoom Telephonics Inc" } ,
    { 0x141c, "Zoom", "Zoom Telephonics, Inc" } ,
    { 0x141E, NULL, "Fanuc Co. Ltd" } ,
    { 0x141F, NULL, "Visiontech Ltd" } ,
    { 0x1420, NULL, "Psion Dacom PLC" } ,
    { 0x1425, NULL, "ASIC Designers Inc" } ,
    { 0x1428, NULL, "Edec Co Ltd" } ,
    { 0x1429, NULL, "Unex Technology Corp." } ,
    { 0x142A, NULL, "Kingmax Technology Inc" } ,
    { 0x142B, NULL, "Radiolan" } ,
    { 0x142C, NULL, "Minton Optic Industry Co Ltd" } ,
    { 0x142D, NULL, "Pixstream Inc" } ,
    { 0x1430, NULL, "ITT Aerospace/Communications Division" } ,
    { 0x1433, NULL, "Eltec Elektronik AG" } ,
    { 0x1435, "RTD-USA", "Real Time Devices USA, Inc." } ,
    { 0x1436, NULL, "CIS Technology Inc" } ,
    { 0x1437, NULL, "Nissin Inc Co" } ,
    { 0x1438, NULL, "Atmel-Dream" } ,
    { 0x143F, NULL, "Lightwell Co Ltd - Zax Division" } ,
    { 0x1441, NULL, "Agie SA." } ,
    { 0x1445, NULL, "Logical Co Ltd" } ,
    { 0x1446, NULL, "Graphin Co. Ltd" } ,
    { 0x1447, NULL, "Aim GMBH" } ,
    { 0x1448, "Alesis", "Alesis Studio" } ,
    { 0x144A, "ADLINK", "ADLINK Technology Inc" } ,
    { 0x144B, "Loronix", "Loronix Information Systems, Inc." } ,
    { 0x144D, NULL, "Samsung Electronics Co Ltd" } ,
    { 0x1450, NULL, "Octave Communications Ind." } ,
    { 0x1451, NULL, "SP3D Chip Design GMBH" } ,
    { 0x1453, NULL, "Mycom Inc" } ,
    { 0x1455, NULL, "Logic Plus PLUS Inc" } ,
    { 0x1458, "Giga-Byte", "Giga-Byte Technologies" } ,
    { 0x145C, NULL, "Cryptek" } ,
    { 0x145F, "Baldor", "Baldor Electric Company" } ,
    { 0x1460, NULL, "Dynarc Inc" } ,
    { 0x1462, NULL, "Micro-Star International Co Ltd" } ,
    { 0x1463, NULL, "Fast Corporation" } ,
    { 0x1464, "ICS", "Interactive Circuits & Systems Ltd" } ,
    { 0x1465, NULL, "GN Nettest Telecom Div." } ,
    { 0x1468, NULL, "Ambit Microsystems Corp." } ,
    { 0x1469, NULL, "Cleveland Motion Controls" } ,
    { 0x146C, NULL, "Ruby Tech Corp." } ,
    { 0x146D, NULL, "Tachyon Inc." } ,
    { 0x146E, NULL, "WMS Gaming" } ,
    { 0x1471, NULL, "Integrated Telecom Express Inc" } ,
    { 0x1473, NULL, "Zapex Technologies Inc" } ,
    { 0x1474, NULL, "Doug Carson & Associates" } ,
    { 0x1477, NULL, "Net Insight" } ,
    { 0x1478, NULL, "Diatrend Corporation" } ,
    { 0x147B, NULL, "Abit Computer Corp." } ,
    { 0x147F, NULL, "Nihon Unisys Ltd." } ,
    { 0x1482, NULL, "Isytec - Integrierte Systemtechnik Gmbh" } ,
    { 0x1483, NULL, "Labway Coporation" } ,
    { 0x1485, NULL, "Erma - Electronic GMBH" } ,
    { 0x1489, NULL, "KYE Systems Corporation" } ,
    { 0x148A, NULL, "Opto 22" } ,
    { 0x148B, NULL, "Innomedialogic Inc." } ,
    { 0x148D, "Digicom", "Digicom Systems Inc." } ,
    { 0x148E, NULL, "OSI Plus Corporation" } ,
    { 0x148F, NULL, "Plant Equipment Inc." } ,
    { 0x1490, NULL, "TC Labs Pty Ltd." } ,
    { 0x1493, NULL, "Maker Communications" } ,
    { 0x1495, NULL, "Tokai Communications Industry Co. Ltd" } ,
    { 0x1496, NULL, "Joytech Computer Co. Ltd." } ,
    { 0x1497, NULL, "SMA Regelsysteme GMBH" } ,
    { 0x1499, NULL, "Micro-Technology Co Ltd" } ,
    { 0x149B, NULL, "Seiko Instruments Inc" } ,
    { 0x149E, NULL, "Mapletree Networks Inc." } ,
    { 0x149F, NULL, "Lectron Co Ltd" } ,
    { 0x14A0, NULL, "Softing GMBH" } ,
    { 0x14A2, NULL, "Millennium Engineering Inc" } ,
    { 0x14A4, NULL, "GVC/BCM Advanced Research" } ,
    { 0x14A5, NULL, "Xionics Document Technologies Inc." } ,
    { 0x14A9, NULL, "Hivertec Inc." } ,
    { 0x14AB, NULL, "Mentor Graphics Corp." } ,
    { 0x14B1, NULL, "Nextcom K.K." } ,
    { 0x14B3, "Xpeed", "Xpeed Inc." } ,
    { 0x14B4, NULL, "Philips Business Electronics B.V." } ,
    { 0x14B6, NULL, "Quantum Data Corp." } ,
    { 0x14B7, "Proxim", "Proxim Inc." } ,
    { 0x14B9, "Aironet", "Aironet Wireless Communication" } ,
    { 0x14BA, NULL, "Internix Inc." } ,
    { 0x14BB, NULL, "Semtech Corporation" } ,
    { 0x14BE, NULL, "L3 Communications" } ,
    { 0x14C1, NULL, "Myricom Inc." } ,
    { 0x14C2, NULL, "DTK Computer" } ,
    { 0x14C4, NULL, "Iwasaki Information Systems Co Ltd" } ,
    { 0x14C5, NULL, "ABB Automation Products AB" } ,
    { 0x14C6, NULL, "Data Race Inc" } ,
    { 0x14C7, "Modtech", "Modular Technology Ltd." } ,
    { 0x14C8, "Turbocomm", "Turbocomm Tech Inc" } ,
    { 0x14C9, NULL, "Odin Telesystems Inc" } ,
    { 0x14CB, NULL, "Billionton Systems Inc./Cadmus Micro Inc" } ,
    { 0x14CD, NULL, "Universal Scientific Ind." } ,
    { 0x14CF, NULL, "TEK Microsystems Inc." } ,
    { 0x14D2, "OX", "Oxford Semiconductor" } ,
    { 0x14D4, "PANACOM", "Panacom Technology Corporation" } ,
    { 0x14D5, NULL, "Nitsuko Corporation" } ,
    { 0x14D6, NULL, "Accusys Inc" } ,
    { 0x14D7, NULL, "Hirakawa Hewtech Corp" } ,
    { 0x14D8, NULL, "Hopf Elektronik GMBH" } ,
    { 0x14D9, NULL, "Alpha Processor Inc" } ,
    { 0x14DB, "Avlab", "Avlab Technology Inc." } ,
    { 0x14DC, "Amplicon", "Amplicon Liveline Inc." } ,
    { 0x14DD, NULL, "Imodl Inc." } ,
    { 0x14DE, NULL, "Applied Integration Corporation" } ,
    { 0x14E3, NULL, "Amtelco" } ,
    { 0x14E4, NULL, "Broadcom Corporation" } ,
    { 0x14EB, NULL, "Seiko Epson Corporation" } ,
    { 0x14EC, NULL, "Acqiris" } ,
    { 0x14ED, NULL, "Datakinetics Ltd" } ,
    { 0x14EF, NULL, "Carry Computer Eng. Co Ltd" } ,
    { 0x14F1, "Conexant", "Conexant Systems, Inc." } ,
    { 0x14F2, "Mobility", "Mobility Electronics, Inc." } ,
    { 0x14F4, NULL, "Tokyo Electronic Industry Co. Ltd." } ,
    { 0x14F5, NULL, "Sopac Ltd" } ,
    { 0x14F6, NULL, "Coyote Technologies LLC" } ,
    { 0x14F7, NULL, "Wolf Technology Inc" } ,
    { 0x14F8, NULL, "Audiocodes Inc" } ,
    { 0x14F9, NULL, "AG Communications" } ,
    { 0x14FB, NULL, "Transas Marine (UK) Ltd" } ,
    { 0x14FC, NULL, "Quadrics Supercomputers World" } ,
    { 0x14FD, NULL, "Japan Computer Industry Inc." } ,
    { 0x14FE, NULL, "Archtek Telecom Corp." } ,
    { 0x14FF, NULL, "Twinhead International Corp." } ,
    { 0x1500, "DELTA", "DELTA Electronics, Inc." } ,
    { 0x1501, NULL, "Banksoft Canada Ltd" } ,
    { 0x1502, NULL, "Mitsubishi Electric Logistics Support Co" } ,
    { 0x1503, NULL, "Kawasaki LSI USA Inc" } ,
    { 0x1504, NULL, "Kaiser Electronics" } ,
    { 0x1506, NULL, "Chameleon Systems Inc" } ,
    { 0x1507, "Htec", "Htec Ltd." } ,
    { 0x1509, NULL, "First International Computer Inc" } ,
    { 0x150B, NULL, "Yamashita Systems Corp" } ,
    { 0x150C, NULL, "Kyopal Co Ltd" } ,
    { 0x150D, NULL, "Warpspped Inc" } ,
    { 0x150E, NULL, "C-Port Corporation" } ,
    { 0x150F, NULL, "Intec GMBH" } ,
    { 0x1510, NULL, "Behavior Tech Computer Corp" } ,
    { 0x1511, NULL, "Centillium Technology Corp" } ,
    { 0x1512, NULL, "Rosun Technologies Inc" } ,
    { 0x1513, NULL, "Raychem" } ,
    { 0x1514, NULL, "TFL LAN Inc" } ,
    { 0x1515, NULL, "ICS Advent" } ,
    { 0x1516, NULL, "Myson Technology Inc" } ,
    { 0x1517, NULL, "Echotek Corporation" } ,
    { 0x1518, NULL, "PEP Modular Computers GMBH" } ,
    { 0x1519, NULL, "Telefon Aktiebolaget LM Ericsson" } ,
    { 0x151A, "Globetek", "Globetek Inc." } ,
    { 0x151B, NULL, "Combox Ltd" } ,
    { 0x151C, NULL, "Digital Audio Labs Inc" } ,
    { 0x151D, NULL, "Fujitsu Computer Products Of America" } ,
    { 0x151E, NULL, "Matrix Corp." } ,
    { 0x151F, NULL, "Topic Semiconductor Corp" } ,
    { 0x1520, NULL, "Chaplet System Inc" } ,
    { 0x1521, NULL, "Bell Corporation" } ,
    { 0x1522, "Mainpine", "Mainpine Limited" } ,
    { 0x1523, NULL, "Music Semiconductors" } ,
    { 0x1524, NULL, "ENE Technology Inc" } ,
    { 0x1525, NULL, "Impact Technologies" } ,
    { 0x1526, NULL, "ISS Inc" } ,
    { 0x1527, NULL, "Solectron" } ,
    { 0x1528, NULL, "Acksys" } ,
    { 0x1529, NULL, "American Microsystems Inc" } ,
    { 0x152A, NULL, "Quickturn Design Systems" } ,
    { 0x152B, NULL, "Flytech Technology Co Ltd" } ,
    { 0x152C, NULL, "Macraigor Systems LLC" } ,
    { 0x152D, NULL, "Quanta Computer Inc" } ,
    { 0x152E, NULL, "Melec Inc" } ,
    { 0x152F, NULL, "Philips - Crypto" } ,
    { 0x1532, NULL, "Echelon Corporation" } ,
    { 0x1533, NULL, "Baltimore" } ,
    { 0x1534, NULL, "Road Corporation" } ,
    { 0x1535, NULL, "Evergreen Technologies Inc" } ,
    { 0x1537, NULL, "Datalex Communcations" } ,
    { 0x1538, NULL, "Aralion Inc." } ,
    { 0x1539, NULL, "Atelier Informatiques et Electronique Et" } ,
    { 0x153A, NULL, "ONO Sokki" } ,
    { 0x153B, NULL, "Terratec Electronic GMBH" } ,
    { 0x153C, NULL, "Antal Electronic" } ,
    { 0x153D, NULL, "Filanet Corporation" } ,
    { 0x153E, NULL, "Techwell Inc" } ,
    { 0x153F, NULL, "MIPS Denmark" } ,
    { 0x1540, NULL, "Provideo Multimedia Co Ltd" } ,
    { 0x1541, NULL, "Telocity Inc." } ,
    { 0x1542, NULL, "Vivid Technology Inc" } ,
    { 0x1543, NULL, "Silicon Laboratories" } ,
    { 0x1544, NULL, "DCM Data Systems" } ,
    { 0x1545, NULL, "Visiontek" } ,
    { 0x1546, NULL, "IOI Technology Corp." } ,
    { 0x1547, NULL, "Mitutoyo Corporation" } ,
    { 0x1548, NULL, "Jet Propulsion Laboratory" } ,
    { 0x1549, "ISS", "Interconnect Systems Solutions" } ,
    { 0x154A, NULL, "Max Technologies Inc." } ,
    { 0x154B, NULL, "Computex Co Ltd" } ,
    { 0x154C, NULL, "Visual Technology Inc." } ,
    { 0x154D, NULL, "PAN International Industrial Corp" } ,
    { 0x154E, NULL, "Servotest Ltd" } ,
    { 0x154F, NULL, "Stratabeam Technology" } ,
    { 0x1550, NULL, "Open Network Co Ltd" } ,
    { 0x1551, NULL, "Smart Electronic Development GMBH" } ,
    { 0x1552, NULL, "Racal Airtech Ltd" } ,
    { 0x1553, NULL, "Chicony Electronics Co Ltd" } ,
    { 0x1554, NULL, "Prolink Microsystems Corp." } ,
    { 0x1556, NULL, "PLD Applications" } ,
    { 0x1557, NULL, "Mediastar Co. Ltd" } ,
    { 0x1558, NULL, "Clevo/Kapok Computer" } ,
    { 0x1559, NULL, "SI Logic Ltd" } ,
    { 0x155A, NULL, "Innomedia Inc" } ,
    { 0x155B, NULL, "Protac International Corp" } ,
    { 0x155C, NULL, "Cemax-Icon Inc" } ,
    { 0x155D, NULL, "MAC System Co Ltd" } ,
    { 0x155E, NULL, "LP Elektronik GMBH" } ,
    { 0x155F, NULL, "Perle Systems Limited" } ,
    { 0x1560, NULL, "Terayon Communications Systems" } ,
    { 0x1561, NULL, "Viewgraphics Inc" } ,
    { 0x1562, NULL, "Symbol Technologies" } ,
    { 0x1563, NULL, "A-Trend Technology Co Ltd" } ,
    { 0x1564, NULL, "Yamakatsu Electronics Industry Co Ltd" } ,
    { 0x1565, NULL, "Biostar Microtech Intl Corp" } ,
    { 0x1566, NULL, "Ardent Technologies Inc" } ,
    { 0x1567, NULL, "Jungsoft" } ,
    { 0x1568, NULL, "DDK Electronics Inc" } ,
    { 0x1569, NULL, "Palit Microsystems Inc" } ,
    { 0x156A, NULL, "Avtec Systems" } ,
    { 0x156B, NULL, "2wire Inc" } ,
    { 0x156C, NULL, "Vidac Electronics GMBH" } ,
    { 0x156D, NULL, "Alpha-Top Corp" } ,
    { 0x156E, NULL, "Alfa Inc." } ,
    { 0x156F, NULL, "M-Systems Flash Disk Pioneers Ltd" } ,
    { 0x1570, NULL, "Lecroy Corporation" } ,
    { 0x1571, NULL, "Contemporary Controls" } ,
    { 0x1572, NULL, "Otis Elevator Company" } ,
    { 0x1573, NULL, "Lattice - Vantis" } ,
    { 0x1574, NULL, "Fairchild Semiconductor" } ,
    { 0x1575, NULL, "Voltaire Advanced Data Security Ltd" } ,
    { 0x1576, NULL, "Viewcast Com" } ,
    { 0x1578, NULL, "Hitt" } ,
    { 0x1579, NULL, "Dual Technology Corporation" } ,
    { 0x157A, NULL, "Japan Elecronics Ind. Inc" } ,
    { 0x157B, NULL, "Star Multimedia Corp." } ,
    { 0x157C, "Eurosoft", "Eurosoft (UK)" } ,
    { 0x157D, NULL, "Gemflex Networks" } ,
    { 0x157E, NULL, "Transition Networks" } ,
    { 0x157F, NULL, "PX Instruments Technology Ltd" } ,
    { 0x1580, NULL, "Primex Aerospace Co." } ,
    { 0x1581, NULL, "SEH Computertechnik GMBH" } ,
    { 0x1582, NULL, "Cytec Corporation" } ,
    { 0x1583, NULL, "Inet Technologies Inc" } ,
    { 0x1584, NULL, "Uniwill Computer Corp." } ,
    { 0x1585, NULL, "Marconi Commerce Systems SRL" } ,
    { 0x1586, NULL, "Lancast Inc" } ,
    { 0x1587, NULL, "Konica Corporation" } ,
    { 0x1588, NULL, "Solidum Systems Corp" } ,
    { 0x1589, NULL, "Atlantek Microsystems Pty Ltd" } ,
    { 0x158A, NULL, "Digalog Systems Inc" } ,
    { 0x158B, NULL, "Allied Data Technologies" } ,
    { 0x158C, NULL, "Hitachi Semiconductor & Devices Sales Co" } ,
    { 0x158D, NULL, "Point Multimedia Systems" } ,
    { 0x158E, NULL, "Lara Technology Inc" } ,
    { 0x158F, NULL, "Ditect Coop" } ,
    { 0x1590, NULL, "3pardata Inc." } ,
    { 0x1591, NULL, "ARN" } ,
    { 0x1592, "Syba", "Syba Tech Ltd." } ,
    { 0x1593, NULL, "Bops Inc" } ,
    { 0x1594, NULL, "Netgame Ltd" } ,
    { 0x1595, NULL, "Diva Systems Corp." } ,
    { 0x1596, NULL, "Folsom Research Inc" } ,
    { 0x1597, NULL, "Memec Design Services" } ,
    { 0x1598, NULL, "Granite Microsystems" } ,
    { 0x1599, NULL, "Delta Electronics Inc" } ,
    { 0x159A, NULL, "General Instrument" } ,
    { 0x159B, NULL, "Faraday Technology Corp" } ,
    { 0x159C, NULL, "Stratus Computer Systems" } ,
    { 0x159D, NULL, "Ningbo Harrison Electronics Co Ltd" } ,
    { 0x159E, NULL, "A-Max Technology Co Ltd" } ,
    { 0x159F, NULL, "Galea Network Security" } ,
    { 0x15A0, NULL, "Compumaster SRL" } ,
    { 0x15A1, NULL, "Geocast Network Systems Inc" } ,
    { 0x15A2, NULL, "Catalyst Enterprises Inc" } ,
    { 0x15A3, NULL, "Italtel" } ,
    { 0x15A4, NULL, "X-Net OY" } ,
    { 0x15A5, NULL, "Toyota MACS Inc" } ,
    { 0x15A6, NULL, "Sunlight Ultrasound Technologies Ltd" } ,
    { 0x15A7, NULL, "SSE Telecom Inc" } ,
    { 0x15A8, NULL, "Shanghai Communications Technologies Cen" } ,
    { 0x15AA, NULL, "Moreton Bay" } ,
    { 0x15AB, NULL, "Bluesteel Networks Inc" } ,
    { 0x15AC, NULL, "North Atlantic Instruments" } ,
    { 0x15AD, "VMware", "VMware Inc." } ,
    { 0x15AE, NULL, "Amersham Pharmacia Biotech" } ,
    { 0x15B0, NULL, "Zoltrix International Limited" } ,
    { 0x15B1, NULL, "Source Technology Inc" } ,
    { 0x15B2, NULL, "Mosaid Technologies Inc." } ,
    { 0x15B3, NULL, "Mellanox Technology" } ,
    { 0x15B4, NULL, "CCI/Triad" } ,
    { 0x15B5, NULL, "Cimetrics Inc" } ,
    { 0x15B6, NULL, "Texas Memory Systems Inc" } ,
    { 0x15B7, NULL, "Sandisk Corp." } ,
    { 0x15B8, NULL, "Addi-Data GMBH" } ,
    { 0x15B9, NULL, "Maestro Digital Communications" } ,
    { 0x15BA, NULL, "Impacct Technology Corp" } ,
    { 0x15BB, NULL, "Portwell Inc" } ,
    { 0x15BC, "Agilent", "Agilent Technologies" } ,
    { 0x15BD, NULL, "DFI Inc." } ,
    { 0x15BE, NULL, "Sola Electronics" } ,
    { 0x15BF, NULL, "High Tech Computer Corp (HTC)" } ,
    { 0x15C0, "BVM", "BVM Limited" } ,
    { 0x15C1, NULL, "Quantel" } ,
    { 0x15C2, NULL, "Newer Technology Inc" } ,
    { 0x15C3, NULL, "Taiwan Mycomp Co Ltd" } ,
    { 0x15C4, NULL, "EVSX Inc" } ,
    { 0x15C5, NULL, "Procomp Informatics Ltd" } ,
    { 0x15C6, NULL, "Technical University Of Budapest" } ,
    { 0x15C7, NULL, "Tateyama System Laboratory Co Ltd" } ,
    { 0x15C8, NULL, "Penta Media Co. Ltd" } ,
    { 0x15C9, NULL, "Serome Technology Inc" } ,
    { 0x15CA, NULL, "Bitboys OY" } ,
    { 0x15CB, NULL, "AG Electronics Ltd" } ,
    { 0x15CC, NULL, "Hotrail Inc." } ,
    { 0x15CD, NULL, "Dreamtech Co Ltd" } ,
    { 0x15CE, NULL, "Genrad Inc." } ,
    { 0x15CF, NULL, "Hilscher GMBH" } ,
    { 0x15D1, "Infineon", "Infineon Technologies AG" } ,
    { 0x15D2, NULL, "FIC (First International Computer Inc)" } ,
    { 0x15D3, NULL, "NDS Technologies Israel Ltd" } ,
    { 0x15D4, NULL, "Iwill Corporation" } ,
    { 0x15D5, NULL, "Tatung Co." } ,
    { 0x15D6, NULL, "Entridia Corporation" } ,
    { 0x15D7, NULL, "Rockwell-Collins Inc" } ,
    { 0x15D8, NULL, "Cybernetics Technology Co Ltd" } ,
    { 0x15D9, NULL, "Super Micro Computer Inc" } ,
    { 0x15DA, NULL, "Cyberfirm Inc." } ,
    { 0x15DB, NULL, "Applied Computing Systems Inc." } ,
    { 0x15DC, "Litronic", "Litronic Inc." } ,
    { 0x15DD, NULL, "Sigmatel Inc." } ,
    { 0x15DE, NULL, "Malleable Technologies Inc" } ,
    { 0x15DF, NULL, "Infinilink Corp." } ,
    { 0x15E0, NULL, "Cacheflow Inc" } ,
    { 0x15E1, "VTG", "Voice Technologies Group" } ,
    { 0x15E2, NULL, "Quicknet Technologies Inc" } ,
    { 0x15E3, NULL, "Networth Technologies Inc" } ,
    { 0x15E4, NULL, "VSN Systemen BV" } ,
    { 0x15E5, NULL, "Valley Technologies Inc" } ,
    { 0x15E6, NULL, "Agere Inc." } ,
    { 0x15E7, NULL, "GET Engineering Corp." } ,
    { 0x15E8, NULL, "National Datacomm Corp." } ,
    { 0x15E9, NULL, "Pacific Digital Corp." } ,
    { 0x15EA, NULL, "Tokyo Denshi Sekei K.K." } ,
    { 0x15EB, NULL, "Drsearch GMBH" } ,
    { 0x15EC, NULL, "Beckhoff GMBH" } ,
    { 0x15ED, NULL, "Macrolink Inc" } ,
    { 0x15EE, NULL, "IN Win Development Inc." } ,
    { 0x15EF, NULL, "Intelligent Paradigm Inc" } ,
    { 0x15F0, NULL, "B-Tree Systems Inc" } ,
    { 0x15F1, NULL, "Times N Systems Inc" } ,
    { 0x15F2, NULL, "Diagnostic Instruments Inc" } ,
    { 0x15F3, NULL, "Digitmedia Corp." } ,
    { 0x15F4, NULL, "Valuesoft" } ,
    { 0x15F5, NULL, "Power Micro Research" } ,
    { 0x15F6, NULL, "Extreme Packet Device Inc" } ,
    { 0x15F7, NULL, "Banctec" } ,
    { 0x15F8, NULL, "Koga Electronics Co" } ,
    { 0x15F9, NULL, "Zenith Electronics Corporation" } ,
    { 0x15FA, NULL, "J.P. Axzam Corporation" } ,
    { 0x15FB, NULL, "Zilog Inc." } ,
    { 0x15FC, NULL, "Techsan Electronics Co Ltd" } ,
    { 0x15FD, NULL, "N-Cubed.Net" } ,
    { 0x15FE, NULL, "Kinpo Electronics Inc" } ,
    { 0x15FF, NULL, "Fastpoint Technologies Inc." } ,
    { 0x1600, NULL, "Northrop Grumman - Canada Ltd" } ,
    { 0x1601, NULL, "Tenta Technology" } ,
    { 0x1602, NULL, "Prosys-TEC Inc." } ,
    { 0x1603, NULL, "Nokia Wireless Business Communications" } ,
    { 0x1604, NULL, "Central System Research Co Ltd" } ,
    { 0x1605, NULL, "Pairgain Technologies" } ,
    { 0x1606, NULL, "Europop AG" } ,
    { 0x1607, NULL, "Lava Semiconductor Manufacturing Inc." } ,
    { 0x1608, NULL, "Automated Wagering International" } ,
    { 0x1609, NULL, "Sciemetric Instruments Inc" } ,
    { 0x160A, NULL, "Kollmorgen Servotronix" } ,
    { 0x160B, NULL, "Onkyo Corp." } ,
    { 0x160C, NULL, "Oregon Micro Systems Inc." } ,
    { 0x160D, NULL, "Aaeon Electronics Inc" } ,
    { 0x160E, NULL, "CML Emergency Services" } ,
    { 0x160F, NULL, "ITEC Co Ltd" } ,
    { 0x1610, NULL, "Tottori Sanyo Electric Co Ltd" } ,
    { 0x1611, NULL, "Bel Fuse Inc." } ,
    { 0x1612, NULL, "Telesynergy Research Inc." } ,
    { 0x1613, NULL, "System Craft Inc." } ,
    { 0x1614, NULL, "Jace Tech Inc." } ,
    { 0x1615, NULL, "Equus Computer Systems Inc" } ,
    { 0x1616, NULL, "Iotech Inc." } ,
    { 0x1617, NULL, "Rapidstream Inc" } ,
    { 0x1618, NULL, "Esec SA" } ,
    { 0x1619, "FarSite", "FarSite Communications Limited" } ,
    { 0x161A, NULL, "Wvinten Ltd" } ,
    { 0x161B, NULL, "Mobilian Israel Ltd" } ,
    { 0x161C, NULL, "Berkshire Products" } ,
    { 0x161D, NULL, "Gatec" } ,
    { 0x161E, NULL, "Kyoei Sangyo Co Ltd" } ,
    { 0x161F, NULL, "Arima Computer Co" } ,
    { 0x1620, NULL, "Sigmacom Co Ltd" } ,
    { 0x1621, NULL, "Lynx Studio Technology Inc" } ,
    { 0x1622, "NHC", "Nokia Home Communications" } ,
    { 0x1623, NULL, "KRF Tech Ltd" } ,
    { 0x1624, NULL, "CE Infosys GMBH" } ,
    { 0x1625, NULL, "Warp Nine Engineering" } ,
    { 0x1626, NULL, "TDK Semiconductor Corp." } ,
    { 0x1627, NULL, "BCom Electronics Inc" } ,
    { 0x1629, NULL, "Kongsberg Spacetec a.s." } ,
    { 0x162A, NULL, "Sejin Computerland Co Ltd" } ,
    { 0x162B, NULL, "Shanghai Bell Company Limited" } ,
    { 0x162C, NULL, "C&H Technologies Inc" } ,
    { 0x162D, NULL, "Reprosoft Co Ltd" } ,
    { 0x162E, NULL, "Margi Systems Inc" } ,
    { 0x162F, NULL, "Rohde & Schwarz GMBH & Co KG" } ,
    { 0x1630, NULL, "Sky Computers Inc" } ,
    { 0x1631, NULL, "NEC Computer International" } ,
    { 0x1632, NULL, "Verisys Inc" } ,
    { 0x1633, NULL, "Adac Corporation" } ,
    { 0x1634, NULL, "Visionglobal Network Corp." } ,
    { 0x1635, NULL, "Decros" } ,
    { 0x1636, NULL, "Jean Company Ltd" } ,
    { 0x1637, NULL, "NSI" } ,
    { 0x1638, NULL, "Eumitcom Technology Inc" } ,
    { 0x163A, NULL, "Air Prime Inc" } ,
    { 0x163B, NULL, "Glotrex Co Ltd" } ,
    { 0x163C, NULL, "Smart Link" } ,
    { 0x163D, NULL, "Heidelberg Digital LLC" } ,
    { 0x163E, NULL, "3dpower" } ,
    { 0x163F, NULL, "Renishaw PLC" } ,
    { 0x1640, NULL, "Intelliworxx Inc" } ,
    { 0x1641, NULL, "MKNet Corporation" } ,
    { 0x1642, NULL, "Bitland" } ,
    { 0x1643, NULL, "Hajime Industries Ltd" } ,
    { 0x1644, NULL, "Western Avionics Ltd" } ,
    { 0x1645, NULL, "Quick-Serv. Computer Co. Ltd" } ,
    { 0x1646, NULL, "Nippon Systemware Co Ltd" } ,
    { 0x1647, NULL, "Hertz Systemtechnik GMBH" } ,
    { 0x1648, NULL, "MeltDown Systems LLC" } ,
    { 0x1649, NULL, "Jupiter Systems" } ,
    { 0x164A, NULL, "Aiwa Co. Ltd" } ,
    { 0x164C, NULL, "Department Of Defense" } ,
    { 0x164D, NULL, "Ishoni Networks" } ,
    { 0x164E, NULL, "Micrel Inc." } ,
    { 0x164F, NULL, "Datavoice (Pty) Ltd." } ,
    { 0x1650, NULL, "Admore Technology Inc." } ,
    { 0x1651, NULL, "Chaparral Network Storage" } ,
    { 0x1652, NULL, "Spectrum Digital Inc." } ,
    { 0x1653, NULL, "Nature Worldwide Technology Corp" } ,
    { 0x1654, NULL, "Sonicwall Inc" } ,
    { 0x1655, NULL, "Dazzle Multimedia Inc." } ,
    { 0x1656, NULL, "Insyde Software Corp" } ,
    { 0x1657, NULL, "Brocade Communications Systems" } ,
    { 0x1658, NULL, "Med Associates Inc." } ,
    { 0x1659, NULL, "Shiba Denshi Systems Inc." } ,
    { 0x165A, NULL, "Epix Inc." } ,
    { 0x165B, NULL, "Real-Time Digital Inc." } ,
    { 0x165C, NULL, "Gidel Ltd." } ,
    { 0x165D, NULL, "Hsing Tech. Enterprise Co. Ltd." } ,
    { 0x165E, NULL, "Hyunju Computer Co. Ltd." } ,
    { 0x165F, NULL, "Add One Company" } ,
    { 0x1660, NULL, "Network Security Technologies Inc. (Net " } ,
    { 0x1661, NULL, "Worldspace Corp." } ,
    { 0x1662, NULL, "Int Labs" } ,
    { 0x1663, NULL, "Elmec Inc. Ltd." } ,
    { 0x1664, NULL, "Fastfame Technology Co. Ltd." } ,
    { 0x1665, NULL, "Edax Inc." } ,
    { 0x1666, NULL, "Norpak Corporation" } ,
    { 0x1667, NULL, "CoSystems Inc." } ,
    { 0x1668, "Actiontec", "Actiontec Electronics Inc." } ,
    { 0x166A, NULL, "Komatsu Ltd." } ,
    { 0x166B, NULL, "Supernet Inc." } ,
    { 0x166C, NULL, "Shade Ltd." } ,
    { 0x166D, NULL, "Sibyte Inc." } ,
    { 0x166E, NULL, "Schneider Automation Inc." } ,
    { 0x166F, NULL, "Televox Software Inc." } ,
    { 0x1670, NULL, "Rearden Steel" } ,
    { 0x1671, NULL, "Atan Technology Inc." } ,
    { 0x1672, NULL, "Unitec Co. Ltd." } ,
    { 0x1673, NULL, "Connex" } ,
    { 0x1675, NULL, "Square Wave Technology" } ,
    { 0x1676, NULL, "Emachines Inc." } ,
    { 0x1677, NULL, "Bernecker + Rainer" } ,
    { 0x1678, NULL, "INH Semiconductor" } ,
    { 0x1679, NULL, "Tokyo Electron Device Ltd." } ,
    { 0x168A, "USA", "Utimaco Safeware AG" } ,
    { 0x16AE, "SFNT", "SafeNet Inc." } ,
    { 0x1755, "Alchemy", "Alchemy Semiconductor Inc." } ,
    { 0x1813, NULL, "Modem Silicon Operation" } ,
    { 0X1898, "DICIT", "DIC INFORMATION TECHNOLOGY ,LTD" } ,
    { 0x1A08, "Sierra", "Sierra Semiconductor" } ,
    { 0x1B13, NULL, "Jaton Corporation" } ,
    { 0x1C1C, "Symphony", "Symphony" } ,
    { 0x1D44, "DPT", "Distributed Processing Technology" } ,
    { 0x1DE1, "Tekram", "Tekram" } ,
    { 0x1DE2, "A/DHOC", "A/DHOC Systmes" } ,
    { 0x1DEA, "LST", "Lasentec" } ,
    { 0x2001, NULL, "Temporal Research Ltd" } ,
    { 0x2002, "AT", "Automation Technology GmbH" } ,
    { 0x2014, "NONTECH", "NONTECH Nonnenmacher GmbH" } ,
    { 0x2348, "Racore", "Racore" } ,
    { 0x2646, NULL, "Kingston Technology Co." } ,
    { 0x270F, "ChainTech", "ChainTech Computer Co. Ltd." } ,
    { 0x2EC1, NULL, "Zenic Inc" } ,
    { 0x3000, "Hansol", "Hansol Electronics Inc." } ,
    { 0x3142, "PostImpressions", "Post Impressions Systems" } ,
    { 0x3388, "Hint", "Hint Corp." } ,
    { 0x3411, NULL, "Quantum Designs (H.K.) Inc." } ,
    { 0x3513, "ARCOM", "ARCOM Control Systems Ltd." } ,
    { 0x38EF, NULL, "4links" } ,
    { 0x3D3D, "3DLabs", "3Dlabs, Inc. Ltd" } ,
    { 0x4005, "Avance", "Avance Logic Inc." } ,
    { 0x4033, "Addtron", "Addtron Technology Co., Inc." } ,
    { 0x4143, "DEC", "Digital Equipment Corp." } ,
    { 0x416C, NULL, "Aladdin Knowledge Systems" } ,
    { 0x4444, "ICompression", "ICompression Inc." } ,
    { 0x4468, "Bridgeport", "Bridgeport Machines" } ,
    { 0x4550, "TDT", "Tucker-Davis Technologies" } ,
    { 0x4594, NULL, "Cogetec Informatique Inc." } ,
    { 0x45FB, "Baldor", "Baldor Electric Company" } ,
    { 0x4680, "UMAX Comp", "UMAX Computer Corp." } ,
    { 0x4843, "Hercules", "Hercules Computer Technology" } ,
    { 0x4943, NULL, "Growth Networks" } ,
    { 0x4954, "Integral", "Integral Technologies" } ,
    { 0x4978, "Axil", "Axil Computer Inc." } ,
    { 0x4A14, "NetVin", "NetVin" } ,
    { 0x4abd, "ABI", "Applied Biosystems Division" } ,
    { 0x4B10, "Buslogic", "Buslogic Inc" } ,
    { 0x4C48, "Lung Hwa", "Lung Hwa Electronics" } ,
    { 0x4C53, NULL, "SBS-OR Industrial Computers" } ,
    { 0x4CA1, NULL, "Seanix Technology Inc" } ,
    { 0x4D51, "Mediaq", "Mediaq Inc." } ,
    { 0x4D54, NULL, "Microtechnica Co Ltd" } ,
    { 0x4DDC, "ILC", "ILC Data Device Corp." } ,
    { 0x5046, "Guillemot", "Guillemot" } ,
    { 0x5053, "TBS/Voyetra", "TBS/Voyetra Technologies" } ,
    { 0x5136, NULL, "S S Technologies" } ,
    { 0x5143, "Qualcomm", "Qualcomm Inc." } ,
    { 0x5145, "ENSONIQ", "ENSONIQ" } ,
    { 0x5301, "Alliance", "Alliance Semicondutor Corp." } ,
    { 0x5333, "S3", "S3 Incorporated" } ,
    { 0x544C, NULL, "Teralogic Inc" } ,
    { 0x5455, "TU-Berlin", "Technische Universtiaet Berlin" } ,
    { 0x5519, "Cnet", "Cnet Technoliges, Inc." } ,
    { 0x5544, "Dunord", "Dunord Technologies" } ,
    { 0x5555, "Genroco", "Genroco Inc." } ,
    { 0x5700, "Netpower", "Netpower" } ,
    { 0x6356, "UltraStor", "UltraStor" } ,
    { 0x6374, "C4T", "c't Magazin f_r Computertechnik" } ,
    { 0x6409, NULL, "Logitec Corp." } ,
    { 0x6666, "Decision", "Decision Computer International Co." } ,
    { 0x7604, "O.N.", "O.N. Electric Co. Ltd." } ,
    { 0x7747, "DaoGuo", "DaoGuo Technology Co.,Ltd" } ,
    { 0x7BDE, "MIDAC", "MIDAC Corporation" } ,
    { 0x7FED, "PowerTV", "PowerTV" } ,
    { 0x8001, "BEYERTONE", "Beyertone AG - Germany" } ,
    { 0x8008, "QUANCOM", "QUANCOM Informationssysteme GmbH" } ,
    { 0x8086, "Intel", "Intel Corporation" } ,
    { 0x8800, "Trigem", "Trigem Computer" } ,
    { 0x8866, "T-Square", "T-Square Design Inc." } ,
    { 0x8888, "Sil Magic", "Silicon Magic" } ,
    { 0x8E0E, "Computone", "Computone Corporation" } ,
    { 0x8E2E, "KTI", "KTI" } ,
    { 0x9004, "Adaptec", "Adaptec" } ,
    { 0x9005, "Adaptec", "Adaptec" } ,
    { 0x907F, "Atronics", "Atronics" } ,
    { 0x919A, NULL, "Gigapixel Corp" } ,
    { 0x9412, "Holtek", "Holtek" } ,
    { 0x9699, NULL, "Omni Media Technology Inc." } ,
    { 0x9710, "Netmos", "Netmos" } ,
    { 0x9877, "surveyors", "pan asia superitendence corp" } ,
    { 0x9902, "StarGen", "StarGen, Inc." } ,
    { 0xA0A0, "Aopen", "Aopen Inc." } ,
    { 0xA0F1, NULL, "Unisys Corporation" } ,
    { 0xA200, "NEC", "NEC Corp." } ,
    { 0xA259, NULL, "Hewlett Packard" } ,
    { 0xA25B, "Hewlett Packard", "Hewlett Packard GmbH PL24-MKT" } ,
    { 0xA304, "Sony", "Sony" } ,
    { 0xA727, NULL, "3com Corporation" } ,
    { 0xAA42, "Scitex", "Scitex Digital Video" } ,
    { 0xAC1E, NULL, "Digital Receiver Technology Inc" } ,
    { 0xAECB, NULL, "Adrienne Electronics Corporation" } ,
    { 0xB00C, "IC Book", "IC Book Labs" } ,
    { 0xB1B3, "Shiva", "Shiva Europe Ltd." } ,
    { 0xB894, NULL, "Brown & Sharpe Mfg. Co." } ,
    { 0xC001, "TSI", "TSI Telsys" } ,
    { 0xC0A9, "Micron/Crucial", "Micron/Crucial Technology" } ,
    { 0xC0DE, NULL, "Motorola" } ,
    { 0xC0FE, "Mot Engrg", "Motion Engineering Inc." } ,
    { 0xC622, NULL, "Hudson Soft Co Ltd" } ,
    { 0xCA50, "Varian", "Varian Australia Pty. Ltd." } ,
    { 0xCAFE, NULL, "Chrysalis-ITS" } ,
    { 0xCCCC, NULL, "Catapult Communications" } ,
    { 0xCDDD, "Tyzx Inc.", "Tyzx Inc." } ,
    { 0xD4D4, "DY4", "DY4 Systems Inc." } ,
    { 0xD84D, "Exsys", "Exsys" } ,
    { 0xDC93, NULL, "Dawicontrol" } ,
    { 0xDEAD, "Indigita", "Indigita Corporation" } ,
    { 0xDEAF, NULL, "Middle Digital, Inc" } ,
    { 0xE000, "Winbond", "Winbond" } ,
    { 0xE159, "Tiger Jet", "Tiger Jet Network Inc" } ,
    { 0xE4BF, NULL, "EKF Elektronik GMBH" } ,
    { 0xEA01, NULL, "Eagle Technology" } ,
    { 0xECC0, "Echo", "Echo Corporation" } ,
    { 0xEDD8, "ARK Logic", "ARK Logic, Inc" } ,
    { 0xF5F5, NULL, "F5 Networks Inc." } ,
    { 0xFA57, "FAST", "Fast Search & Transfer ASA" } ,
    { 0xFEDA, NULL, "Epigram Inc"} ,
    { 0xFFFE, "VMware", "VMware Inc."} ,
    { 0x16da, "MarekMicro", "MarekMicro GmbH"},
    { 0xFFFF, NULL, NULL}
};

#endif
